# Machine manipulation request handlers
#
# Copyright (C) 2020 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import sys
import os
import re
import pathlib
import logging
import asyncio
import platform
import socket
import ipaddress
import time
import shutil
import distro
import tempfile
import getpass
import configparser
from ..confighelper import FileSourceWrapper
from ..utils import source_info, cansocket, sysfs_devs, load_system_module
from ..utils import json_wrapper as jsonw
from ..common import RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    cast
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .application import MoonrakerApp
    from .klippy_connection import KlippyConnection
    from .http_client import HttpClient
    from .shell_command import ShellCommandFactory as SCMDComp
    from .database import MoonrakerDatabase
    from .file_manager.file_manager import FileManager
    from .announcements import Announcements
    from .proc_stats import ProcStats
    from .dbus_manager import DbusManager
    from dbus_next.aio.proxy_object import ProxyInterface
    from dbus_next.signature import Variant
    SudoReturn = Union[Awaitable[Tuple[str, bool]], Tuple[str, bool]]
    SudoCallback = Callable[[], SudoReturn]

CGROUP_PATH = "/proc/1/cgroup"
SCHED_PATH = "/proc/1/sched"
SYSTEMD_PATH = "/etc/systemd/system"
SD_CID_PATH = "/sys/block/mmcblk0/device/cid"
SD_CSD_PATH = "/sys/block/mmcblk0/device/csd"
SD_MFGRS = {
    '1b': "Samsung",
    '03': "Sandisk",
    '74': "PNY"
}
IP_FAMILIES = {'inet': 'ipv4', 'inet6': 'ipv6'}
NETWORK_UPDATE_SEQUENCE = 10
SERVICE_PROPERTIES = [
    "Requires", "After", "SupplementaryGroups", "EnvironmentFiles",
    "ExecStart", "WorkingDirectory", "FragmentPath", "Description",
    "User"
]
USB_IDS_URL = "http://www.linux-usb.org/usb.ids"

class Machine:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self._allowed_services: List[str] = []
        self._init_allowed_services()
        dist_info: Dict[str, Any]
        dist_info = {'name': distro.name(pretty=True)}
        dist_info.update(distro.info())
        dist_info['release_info'] = distro.distro_release_info()
        dist_info['kernel_version'] = platform.release()
        self.inside_container = False
        self.moonraker_service_info: Dict[str, Any] = {}
        self.sudo_req_lock = asyncio.Lock()
        self.periph_lock = asyncio.Lock()
        self._sudo_password: Optional[str] = None
        sudo_template = config.gettemplate("sudo_password", None)
        if sudo_template is not None:
            self._sudo_password = sudo_template.render()
        self._public_ip = ""
        self.system_info: Dict[str, Any] = {
            'python': {
                "version": tuple(sys.version_info),
                "version_string": sys.version.replace("\n", " ")
            },
            'cpu_info': self._get_cpu_info(),
            'sd_info': self._get_sdcard_info(),
            'distribution': dist_info,
            'virtualization': self._check_inside_container(),
            'network': {},
            'canbus': {}
        }
        self._update_log_rollover(log=True)
        providers: Dict[str, type] = {
            "none": BaseProvider,
            "systemd_cli": SystemdCliProvider,
            "systemd_dbus": SystemdDbusProvider,
            "supervisord_cli": SupervisordCliProvider
        }
        self.provider_type = config.get('provider', 'systemd_dbus')
        pclass = providers.get(self.provider_type)
        if pclass is None:
            raise config.error(f"Invalid Provider: {self.provider_type}")
        self.sys_provider: BaseProvider = pclass(config)
        self.system_info["provider"] = self.provider_type
        logging.info(f"Using System Provider: {self.provider_type}")
        self.validator = InstallValidator(config)
        self.sudo_requests: List[Tuple[SudoCallback, str]] = []

        self.server.register_endpoint(
            "/machine/reboot", RequestType.POST, self._handle_machine_request
        )
        self.server.register_endpoint(
            "/machine/shutdown", RequestType.POST, self._handle_machine_request
        )
        self.server.register_endpoint(
            "/machine/services/restart", RequestType.POST, self._handle_service_request
        )
        self.server.register_endpoint(
            "/machine/services/stop", RequestType.POST, self._handle_service_request
        )
        self.server.register_endpoint(
            "/machine/services/start", RequestType.POST, self._handle_service_request
        )
        self.server.register_endpoint(
            "/machine/system_info", RequestType.GET, self._handle_sysinfo_request
        )
        self.server.register_endpoint(
            "/machine/sudo/info", RequestType.GET, self._handle_sudo_info
        )
        self.server.register_endpoint(
            "/machine/sudo/password", RequestType.POST, self._set_sudo_password
        )
        self.server.register_endpoint(
            "/machine/peripherals/serial", RequestType.GET, self._handle_serial_request
        )
        self.server.register_endpoint(
            "/machine/peripherals/usb", RequestType.GET, self._handle_usb_request
        )
        self.server.register_endpoint(
            "/machine/peripherals/canbus", RequestType.GET, self._handle_can_query
        )
        self.server.register_endpoint(
            "/machine/peripherals/video", RequestType.GET, self._handle_video_request
        )

        self.server.register_notification("machine:service_state_changed")
        self.server.register_notification("machine:sudo_alert")

        # Register remote methods
        self.server.register_remote_method(
            "shutdown_machine", self.sys_provider.shutdown)
        self.server.register_remote_method(
            "reboot_machine", self.sys_provider.reboot)

        # IP network shell commands
        shell_cmd: SCMDComp = self.server.load_component(
            config, 'shell_command')
        self.addr_cmd = shell_cmd.build_shell_command("ip -json -det address")
        iwgetbin = "/sbin/iwgetid"
        if not pathlib.Path(iwgetbin).exists():
            iwgetbin = "iwgetid"
        self.iwgetid_cmd = shell_cmd.build_shell_command(iwgetbin)
        self.init_evt = asyncio.Event()
        self.libcam = self._try_import_libcamera()

    def _init_allowed_services(self) -> None:
        app_args = self.server.get_app_args()
        data_path = app_args["data_path"]
        fpath = pathlib.Path(data_path).joinpath("moonraker.asvc")
        fm: FileManager = self.server.lookup_component("file_manager")
        fm.add_reserved_path("allowed_services", fpath, False)
        default_svcs = source_info.read_asset("default_allowed_services") or ""
        try:
            if not fpath.exists():
                fpath.write_text(default_svcs)
            data = fpath.read_text()
        except Exception:
            logging.exception("Failed to read moonraker.asvc")
            data = default_svcs
        svcs = [svc.strip() for svc in data.split("\n") if svc.strip()]
        for svc in svcs:
            if svc.endswith(".service"):
                svc = svc.rsplit(".", 1)[0]
            if svc not in self._allowed_services:
                self._allowed_services.append(svc)

    def _update_log_rollover(self, log: bool = False) -> None:
        sys_info_msg = "\nSystem Info:"
        for header, info in self.system_info.items():
            sys_info_msg += f"\n\n***{header}***"
            if not isinstance(info, dict):
                sys_info_msg += f"\n  {repr(info)}"
            else:
                for key, val in info.items():
                    sys_info_msg += f"\n  {key}: {val}"
        sys_info_msg += "\n\n***Allowed Services***"
        for svc in self._allowed_services:
            sys_info_msg += f"\n  {svc}"
        self.server.add_log_rollover_item('system_info', sys_info_msg, log=log)

    def _try_import_libcamera(self) -> Any:
        try:
            libcam = load_system_module("libcamera")
            cmgr = libcam.CameraManager.singleton()
            self.server.add_log_rollover_item(
                "libcamera",
                f"Found libcamera Python module, version: {cmgr.version}"
            )
            return libcam
        except Exception:
            if self.server.is_verbose_enabled():
                logging.exception("Failed to import libcamera")
            self.server.add_log_rollover_item(
                "libcamera", "Module libcamera unavailble, import failed"
            )
            return None

    @property
    def public_ip(self) -> str:
        return self._public_ip

    @property
    def unit_name(self) -> str:
        svc_info = self.moonraker_service_info
        unit_name = svc_info.get("unit_name", "moonraker.service")
        return unit_name.split(".", 1)[0]

    def is_service_allowed(self, service: str) -> bool:
        return (
            service in self._allowed_services or
            re.match(r"moonraker[_-]?\d*", service) is not None or
            re.match(r"klipper[_-]?\d*", service) is not None
        )

    def validation_enabled(self) -> bool:
        return self.validator.validation_enabled

    def get_system_provider(self):
        return self.sys_provider

    def is_inside_container(self):
        return self.inside_container

    def get_provider_type(self):
        return self.provider_type

    def get_moonraker_service_info(self):
        return dict(self.moonraker_service_info)

    async def wait_for_init(
        self, timeout: Optional[float] = None
    ) -> None:
        try:
            await asyncio.wait_for(self.init_evt.wait(), timeout)
        except asyncio.TimeoutError:
            pass

    async def component_init(self) -> None:
        eventloop = self.server.get_event_loop()
        eventloop.create_task(self.update_usb_ids())
        await self.validator.validation_init()
        await self.sys_provider.initialize()
        if not self.inside_container:
            virt_info = await self.sys_provider.check_virt_status()
            self.system_info['virtualization'] = virt_info
        await self._parse_network_interfaces(0, notify=False)
        pstats: ProcStats = self.server.lookup_component('proc_stats')
        pstats.register_stat_callback(self._parse_network_interfaces)
        available_svcs = self.sys_provider.get_available_services()
        avail_list = list(available_svcs.keys())
        self.system_info['available_services'] = avail_list
        self.system_info['service_state'] = available_svcs
        svc_info = await self.sys_provider.extract_service_info(
            "moonraker", os.getpid()
        )
        self.moonraker_service_info = svc_info
        self.log_service_info(svc_info)
        self.init_evt.set()

    async def validate_installation(self) -> bool:
        return await self.validator.perform_validation()

    async def on_exit(self) -> None:
        await self.validator.remove_announcement()

    async def _handle_machine_request(self, web_request: WebRequest) -> str:
        ep = web_request.get_endpoint()
        if self.inside_container:
            virt_id = self.system_info['virtualization'].get(
                'virt_identifier', "none")
            raise self.server.error(
                f"Cannot {ep.split('/')[-1]} from within a "
                f"{virt_id} container")
        if ep == "/machine/shutdown":
            await self.sys_provider.shutdown()
        elif ep == "/machine/reboot":
            await self.sys_provider.reboot()
        else:
            raise self.server.error("Unsupported machine request")
        return "ok"

    async def do_service_action(self,
                                action: str,
                                service_name: str
                                ) -> None:
        await self.sys_provider.do_service_action(action, service_name)

    def restart_moonraker_service(self):
        async def wrapper():
            try:
                await self.do_service_action("restart", self.unit_name)
            except Exception:
                pass
        self.server.get_event_loop().create_task(wrapper())

    async def _handle_service_request(self, web_request: WebRequest) -> str:
        name: str = web_request.get_str('service')
        action = web_request.get_endpoint().split('/')[-1]
        if name == self.unit_name:
            if action != "restart":
                raise self.server.error(
                    f"Service action '{action}' not available for moonraker")
            self.restart_moonraker_service()
        elif self.sys_provider.is_service_available(name):
            await self.do_service_action(action, name)
        else:
            if name in self._allowed_services:
                raise self.server.error(f"Service '{name}' not installed")
            raise self.server.error(
                f"Service '{name}' not allowed")
        return "ok"

    async def _handle_sysinfo_request(self,
                                      web_request: WebRequest
                                      ) -> Dict[str, Any]:
        kconn: KlippyConnection
        kconn = self.server.lookup_component("klippy_connection")
        sys_info = self.system_info.copy()
        sys_info["instance_ids"] = {
            "moonraker": self.unit_name,
            "klipper": kconn.unit_name
        }
        return {"system_info": sys_info}

    async def _set_sudo_password(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        async with self.sudo_req_lock:
            self._sudo_password = web_request.get_str("password")
            if not await self.check_sudo_access():
                self._sudo_password = None
                raise self.server.error(
                    "Invalid password, sudo access was denied"
                )
            sudo_responses = ["Sudo password successfully set."]
            restart: bool = False
            failed: List[Tuple[SudoCallback, str]] = []
            failed_msgs: List[str] = []
            if self.sudo_requests:
                while self.sudo_requests:
                    cb, msg = self.sudo_requests.pop(0)
                    try:
                        ret = cb()
                        if isinstance(ret, Awaitable):
                            ret = await ret
                        msg, need_restart = ret
                        sudo_responses.append(msg)
                        restart |= need_restart
                    except asyncio.CancelledError:
                        raise
                    except Exception as e:
                        failed.append((cb, msg))
                        failed_msgs.append(str(e))
                restart = False if len(failed) > 0 else restart
                self.sudo_requests = failed
                if not restart and len(sudo_responses) > 1:
                    # at least one successful response and not restarting
                    eventloop = self.server.get_event_loop()
                    eventloop.delay_callback(
                        .05, self.server.send_event,
                        "machine:sudo_alert",
                        {
                            "sudo_requested": self.sudo_requested,
                            "request_messages": self.sudo_request_messages
                        }
                    )
                if failed_msgs:
                    err_msg = "\n".join(failed_msgs)
                    raise self.server.error(err_msg, 500)
                if restart:
                    self.restart_moonraker_service()
                    sudo_responses.append(
                        "Moonraker is currently in the process of restarting."
                    )
        return {
            "sudo_responses": sudo_responses,
            "is_restarting": restart
        }

    async def _handle_sudo_info(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        check_access = web_request.get_boolean("check_access", False)
        has_sudo: Optional[bool] = None
        if check_access:
            has_sudo = await self.check_sudo_access()
        return {
            "sudo_access": has_sudo,
            "linux_user": self.linux_user,
            "sudo_requested": self.sudo_requested,
            "request_messages": self.sudo_request_messages
        }

    async def _handle_serial_request(self, web_request: WebRequest) -> Dict[str, Any]:
        return {
            "serial_devices": await self.detect_serial_devices()
        }

    async def _handle_usb_request(self, web_request: WebRequest) -> Dict[str, Any]:
        return {
            "usb_devices": await self.detect_usb_devices()
        }

    async def _handle_can_query(self, web_request: WebRequest) -> Dict[str, Any]:
        interface = web_request.get_str("interface", "can0")
        return {
            "can_uuids": await self.query_can_uuids(interface)
        }

    async def _handle_video_request(self, web_request: WebRequest) -> Dict[str, Any]:
        return await self.detect_video_devices()

    def get_system_info(self) -> Dict[str, Any]:
        return self.system_info

    @property
    def sudo_password(self) -> Optional[str]:
        return self._sudo_password

    @sudo_password.setter
    def sudo_password(self, pwd: Optional[str]) -> None:
        self._sudo_password = pwd

    @property
    def sudo_requested(self) -> bool:
        return len(self.sudo_requests) > 0

    @property
    def linux_user(self) -> str:
        return getpass.getuser()

    @property
    def sudo_request_messages(self) -> List[str]:
        return [req[1] for req in self.sudo_requests]

    def register_sudo_request(
        self, callback: SudoCallback, message: str
    ) -> None:
        self.sudo_requests.append((callback, message))
        self.server.send_event(
            "machine:sudo_alert",
            {
                "sudo_requested": True,
                "request_messages": self.sudo_request_messages
            }
        )

    async def check_sudo_access(self, cmds: List[str] = []) -> bool:
        if not cmds:
            cmds = ["systemctl --version", "ls /root"]
        shell_cmd: SCMDComp = self.server.lookup_component("shell_command")
        for cmd in cmds:
            try:
                await self.exec_sudo_command(cmd, timeout=10.)
            except shell_cmd.error:
                return False
        return True

    async def exec_sudo_command(
        self, command: str, tries: int = 1, timeout=2.
    ) -> str:
        proc_input = None
        full_cmd = f"sudo {command}"
        if self._sudo_password is not None:
            proc_input = self._sudo_password
            full_cmd = f"sudo -S {command}"
        shell_cmd: SCMDComp = self.server.lookup_component("shell_command")
        return await shell_cmd.exec_cmd(
            full_cmd, proc_input=proc_input, log_complete=False, attempts=tries,
            timeout=timeout
        )

    def _get_sdcard_info(self) -> Dict[str, Any]:
        sd_info: Dict[str, Any] = {}
        cid_file = pathlib.Path(SD_CID_PATH)
        if not cid_file.exists():
            # No SDCard detected at mmcblk0
            return {}
        try:
            cid_text = cid_file.read_text().strip().lower()
            mid = cid_text[:2]
            sd_info['manufacturer_id'] = mid
            sd_info['manufacturer'] = SD_MFGRS.get(mid, "Unknown")
            sd_info['oem_id'] = cid_text[2:6]
            sd_info['product_name'] = bytes.fromhex(cid_text[6:16]).decode(
                encoding="ascii", errors="ignore")
            sd_info['product_revision'] = \
                f"{int(cid_text[16], 16)}.{int(cid_text[17], 16)}"
            sd_info['serial_number'] = cid_text[18:26]
            mfg_year = int(cid_text[27:29], 16) + 2000
            mfg_month = int(cid_text[29], 16)
            sd_info['manufacturer_date'] = f"{mfg_month}/{mfg_year}"
        except Exception:
            logging.info("Error reading SDCard CID Register")
            return {}
        sd_info['capacity'] = "Unknown"
        sd_info['total_bytes'] = 0
        csd_file = pathlib.Path(SD_CSD_PATH)
        # Read CSD Register
        try:
            csd_reg = bytes.fromhex(csd_file.read_text().strip())
            csd_type = (csd_reg[0] >> 6) & 0x3
            if csd_type == 0:
                # Standard Capacity (CSD Version 1.0)
                max_block_len: int = 2**(csd_reg[5] & 0xF)
                c_size = ((csd_reg[6] & 0x3) << 10) | (csd_reg[7] << 2) | \
                    ((csd_reg[8] >> 6) & 0x3)
                c_mult_reg = ((csd_reg[9] & 0x3) << 1) | (csd_reg[10] >> 7)
                c_mult = 2**(c_mult_reg + 2)
                total_bytes: int = (c_size + 1) * c_mult * max_block_len
                sd_info['capacity'] = f"{(total_bytes / (1024.0**2)):.1f} MiB"
            elif csd_type == 1:
                # High Capacity (CSD Version 2.0)
                c_size = ((csd_reg[7] & 0x3F) << 16) | (csd_reg[8] << 8) | \
                    csd_reg[9]
                total_bytes = (c_size + 1) * 512 * 1024
                sd_info['capacity'] = f"{(total_bytes / (1024.0**3)):.1f} GiB"
            elif csd_type == 2:
                # Ultra Capacity (CSD Version 3.0)
                c_size = ((csd_reg[6]) & 0xF) << 24 | (csd_reg[7] << 16) | \
                    (csd_reg[8] << 8) | csd_reg[9]
                total_bytes = (c_size + 1) * 512 * 1024
                sd_info['capacity'] = f"{(total_bytes / (1024.0**4)):.1f} TiB"
            else:
                # Invalid CSD, skip capacity check
                return sd_info
            sd_info['total_bytes'] = total_bytes
        except Exception:
            logging.info("Error Reading SDCard CSD Register")
        return sd_info

    def _get_cpu_info(self) -> Dict[str, Any]:
        cpu_file = pathlib.Path("/proc/cpuinfo")
        mem_file = pathlib.Path("/proc/meminfo")
        cpu_info = {
            'cpu_count': os.cpu_count(),
            'bits': platform.architecture()[0],
            'processor': platform.processor() or platform.machine(),
            'cpu_desc': "",
            'serial_number': "",
            'hardware_desc': "",
            'model': "",
            'total_memory': None,
            'memory_units': ""
        }
        if cpu_file.exists():
            try:
                cpu_text = cpu_file.read_text().strip()
                cpu_items = [item.strip() for item in cpu_text.split("\n\n")
                             if item.strip()]
                for item in cpu_items:
                    cpu_desc_match = re.search(r"model name\s+:\s+(.+)", item)
                    if cpu_desc_match is not None:
                        cpu_info['cpu_desc'] = cpu_desc_match.group(1).strip()
                        break
                hw_match = re.search(r"Hardware\s+:\s+(.+)", cpu_items[-1])
                if hw_match is not None:
                    cpu_info['hardware_desc'] = hw_match.group(1).strip()
                sn_match = re.search(r"Serial\s+:\s+0*(.+)", cpu_items[-1])
                if sn_match is not None:
                    cpu_info['serial_number'] = sn_match.group(1).strip()
                model_match = re.search(r"Model\s+:\s+(.+)", cpu_items[-1])
                if model_match is not None:
                    cpu_info['model'] = model_match.group(1).strip()
            except Exception:
                logging.info("Error Reading /proc/cpuinfo")
        if mem_file.exists():
            try:
                mem_text = mem_file.read_text().strip()
                for line in mem_text.split('\n'):
                    line = line.strip()
                    if line.startswith("MemTotal:"):
                        parts = line.split()
                        cpu_info['total_memory'] = int(parts[1])
                        cpu_info['memory_units'] = parts[2]
                        break
            except Exception:
                logging.info("Error Reading /proc/meminfo")
        return cpu_info

    def _check_inside_container(self) -> Dict[str, Any]:
        cgroup_file = pathlib.Path(CGROUP_PATH)
        virt_type = virt_id = "none"
        if cgroup_file.exists():
            try:
                data = cgroup_file.read_text()
                container_types = ["docker", "lxc"]
                for ct in container_types:
                    if ct in data:
                        self.inside_container = True
                        virt_type = "container"
                        virt_id = ct
                        logging.info(
                            f"Container detected via cgroup: {ct}"
                        )
                        break
            except Exception:
                logging.exception(f"Error reading {CGROUP_PATH}")

        # Fall back to process schedule check
        if not self.inside_container:
            sched_file = pathlib.Path(SCHED_PATH)
            if sched_file.exists():
                try:
                    data = sched_file.read_text().strip()
                    proc_name = data.split('\n')[0].split()[0]
                    if proc_name not in ["init", "systemd"]:
                        self.inside_container = True
                        virt_type = "container"
                        virt_id = "lxc"
                        if (
                            os.path.exists("/.dockerenv") or
                            os.path.exists("/.dockerinit")
                        ):
                            virt_id = "docker"
                        logging.info(
                            f"Container detected via sched: {virt_id}"
                        )
                except Exception:
                    logging.exception(f"Error reading {SCHED_PATH}")
        return {
            'virt_type': virt_type,
            'virt_identifier': virt_id
        }

    async def _parse_network_interfaces(self,
                                        sequence: int,
                                        notify: bool = True
                                        ) -> None:
        if sequence % NETWORK_UPDATE_SEQUENCE:
            return
        network: Dict[str, Any] = {}
        canbus: Dict[str, Any] = {}
        try:
            # get network interfaces
            resp = await self.addr_cmd.run_with_response(log_complete=False)
            decoded: List[Dict[str, Any]] = jsonw.loads(resp)
            for interface in decoded:
                if interface['operstate'] != "UP":
                    continue
                if interface['link_type'] == "can":
                    infodata: dict = interface.get(
                        "linkinfo", {}).get("info_data", {})
                    canbus[interface['ifname']] = {
                        'tx_queue_len': interface['txqlen'],
                        'bitrate': infodata.get("bittiming", {}).get(
                            "bitrate", -1
                        ),
                        'driver': infodata.get("bittiming_const", {}).get(
                            "name", "unknown"
                        )
                    }
                elif (
                    interface['link_type'] == "ether" and
                    'address' in interface
                ):
                    addresses: List[Dict[str, Any]] = [
                        {
                            'family': IP_FAMILIES[addr['family']],
                            'address': addr['local'],
                            'is_link_local': addr.get('scope', "") == "link"
                        }
                        for addr in interface.get('addr_info', [])
                        if 'family' in addr and 'local' in addr
                    ]
                    if not addresses:
                        continue
                    network[interface['ifname']] = {
                        'mac_address': interface['address'],
                        'ip_addresses': addresses
                    }
        except Exception:
            logging.exception("Error processing network update")
            return
        prev_network = self.system_info.get('network', {})
        if network != prev_network:
            self._find_public_ip()
            if notify:
                self.server.send_event("machine:net_state_changed", network)
        self.system_info['network'] = network
        self.system_info['canbus'] = canbus

    async def get_public_network(self) -> Dict[str, Any]:
        wifis = await self._get_wifi_interfaces()
        public_intf = self._find_public_interface()
        ifname = public_intf["ifname"]
        is_wifi = ifname in wifis
        public_intf["is_wifi"] = is_wifi
        if is_wifi:
            public_intf["ssid"] = wifis[ifname]
        # TODO: Can  we detect the private top level domain? That
        # would be ideal
        public_intf["hostname"] = socket.gethostname()
        return public_intf

    def _find_public_interface(self) -> Dict[str, Any]:
        src_ip = self._find_public_ip()
        networks = self.system_info.get("network", {})
        for ifname, ifinfo in networks.items():
            for addrinfo in ifinfo["ip_addresses"]:
                if addrinfo["is_link_local"]:
                    continue
                fam = addrinfo["family"]
                addr = addrinfo["address"]
                if fam == "ipv6" and not src_ip:
                    ip = ipaddress.ip_address(addr)
                    if ip.is_global:
                        return {
                            "ifname": ifname,
                            "address": addr,
                            "family": fam
                        }
                elif src_ip == addr:
                    return {
                        "ifname": ifname,
                        "address": addr,
                        "family": fam
                    }
        return {
            "ifname": "",
            "address": src_ip or "",
            "family": ""
        }

    def _find_public_ip(self) -> str:
        #  Check for an IPv4 Source IP
        # NOTE: It should also be possible to extract this from
        # the routing table, ie: ip -json route
        # It would be an entry with a "gateway" with the lowest
        # metric.  Might also be able to get IPv6 info from this.
        # However, it would be better to use NETLINK for this rather
        # than run another shell command
        src_ip: str = ""
        # First attempt: use "broadcast" to find the local IP
        addr_info = [
            ("<broadcast>", 0, socket.AF_INET),
            ("10.255.255.255", 1, socket.AF_INET),
            ("2001:db8::1234", 1, socket.AF_INET6),
        ]
        for (addr, port, fam) in addr_info:
            s = socket.socket(fam, socket.SOCK_DGRAM | socket.SOCK_NONBLOCK)
            try:
                if addr == "<broadcast>":
                    s.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                s.connect((addr, port))
                src_ip = s.getsockname()[0]
            except Exception:
                continue
            logging.info(f"Detected Local IP: {src_ip}")
            break
        if src_ip != self._public_ip:
            self._public_ip = src_ip
            self.server.send_event("machine:public_ip_changed", src_ip)
        if not src_ip:
            logging.info("Failed to detect local IP address")
        return src_ip

    async def _get_wifi_interfaces(self) -> Dict[str, Any]:
        # get wifi interfaces
        shell_cmd: SCMDComp = self.server.lookup_component('shell_command')
        wifi_intfs: Dict[str, Any] = {}
        try:
            resp = await self.iwgetid_cmd.run_with_response(log_complete=False)
        except shell_cmd.error:
            logging.info("Failed to run 'iwgetid' command")
            return {}
        if resp:
            for line in resp.split("\n"):
                parts = line.strip().split(maxsplit=1)
                wifi_intfs[parts[0]] = parts[1].split(":")[-1].strip('"')
        return wifi_intfs

    def log_service_info(self, svc_info: Dict[str, Any]) -> None:
        if not svc_info:
            return
        name = svc_info.get("unit_name", "unknown")
        manager = svc_info.get("manager", "systemd").capitalize()
        msg = f"\n{manager} unit {name}:"
        for key, val in svc_info.items():
            if key == "properties":
                msg += "\nProperties:"
                for prop_key, prop in val.items():
                    msg += f"\n**{prop_key}={prop}"
            else:
                msg += f"\n{key}: {val}"
        self.server.add_log_rollover_item(name, msg)

    async def update_usb_ids(self, force: bool = False) -> None:
        async with self.periph_lock:
            db: MoonrakerDatabase = self.server.lookup_component("database")
            client: HttpClient = self.server.lookup_component("http_client")
            dpath = pathlib.Path(self.server.get_app_arg("data_path"))
            usb_ids_path = pathlib.Path(dpath).joinpath("misc/usb.ids")
            if usb_ids_path.is_file() and not force:
                return
            usb_id_req_info: Dict[str, str]
            usb_id_req_info = await db.get_item("moonraker", "usb_id_req_info", {})
            etag: Optional[str] = usb_id_req_info.pop("etag", None)
            last_modified: Optional[str] = usb_id_req_info.pop("last_modified", None)
            headers = {"Accept": "text/plain"}
            if etag is not None and usb_ids_path.is_file():
                headers["If-None-Match"] = etag
            if last_modified is not None and usb_ids_path.is_file():
                headers["If-Modified-Since"] = last_modified
            logging.info("Fetching latest usb.ids file...")
            resp = await client.get(
                USB_IDS_URL, headers, enable_cache=False
            )
            if resp.has_error():
                logging.info("Failed to retrieve usb.ids file")
                return
            if resp.status_code == 304:
                logging.info("USB IDs file up to date")
                return
            # Save etag and modified headers
            if resp.etag is not None:
                usb_id_req_info["etag"] = resp.etag
            if resp.last_modified is not None:
                usb_id_req_info["last_modifed"] = resp.last_modified
            await db.insert_item("moonraker", "usb_id_req_info", usb_id_req_info)
            # Write file
            logging.info("Writing usb.ids file...")
            eventloop = self.server.get_event_loop()
            await eventloop.run_in_thread(usb_ids_path.write_bytes, resp.content)

    async def detect_serial_devices(self) -> List[Dict[str, Any]]:
        async with self.periph_lock:
            eventloop = self.server.get_event_loop()
            return await eventloop.run_in_thread(sysfs_devs.find_serial_devices)

    async def detect_usb_devices(self) -> List[Dict[str, Any]]:
        async with self.periph_lock:
            eventloop = self.server.get_event_loop()
            return await eventloop.run_in_thread(self._do_usb_detect)

    def _do_usb_detect(self) -> List[Dict[str, Any]]:
        data_path = pathlib.Path(self.server.get_app_args()["data_path"])
        usb_id_path = data_path.joinpath("misc/usb.ids")
        usb_id_data = sysfs_devs.UsbIdData(usb_id_path)
        dev_list = sysfs_devs.find_usb_devices()
        for usb_dev_info in dev_list:
            cls_ids: List[str] = usb_dev_info.pop("class_ids", None)
            class_info = usb_id_data.get_class_info(*cls_ids)
            usb_dev_info.update(class_info)
            prod_info = usb_id_data.get_product_info(
                usb_dev_info["vendor_id"], usb_dev_info["product_id"]
            )
            for field, desc in prod_info.items():
                if usb_dev_info.get(field) is None:
                    usb_dev_info[field] = desc
        return dev_list

    async def query_can_uuids(self, interface: str) -> List[Dict[str, Any]]:
        async with self.periph_lock:
            cansock = cansocket.CanSocket(interface)
            uuids = await cansocket.query_klipper_uuids(cansock)
            cansock.close()
        return uuids

    async def detect_video_devices(self) -> Dict[str, List[Dict[str, Any]]]:
        async with self.periph_lock:
            eventloop = self.server.get_event_loop()
            v4l2_devs = await eventloop.run_in_thread(sysfs_devs.find_video_devices)
            libcam_devs = await eventloop.run_in_thread(self.get_libcamera_devices)
        return {
            "v4l2_devices": v4l2_devs,
            "libcamera_devices": libcam_devs
        }

    def get_libcamera_devices(self) -> List[Dict[str, Any]]:
        libcam = self.libcam
        libcam_devs: List[Dict[str, Any]] = []
        if libcam is not None:
            cm = libcam.CameraManager.singleton()
            for cam in cm.cameras:
                device: Dict[str, Any] = {"libcamera_id": cam.id}
                props_by_name = {cid.name: val for cid, val in cam.properties.items()}
                device["model"] = props_by_name.get("Model")
                modes: List[Dict[str, Any]] = []
                cam_config = cam.generate_configuration([libcam.StreamRole.Raw])
                for stream_cfg in cam_config:
                    formats = stream_cfg.formats
                    for pix_fmt in formats.pixel_formats:
                        cur_mode: Dict[str, Any] = {"format": str(pix_fmt)}
                        resolutions: List[str] = []
                        for size in formats.sizes(pix_fmt):
                            resolutions.append(str(size))
                        cur_mode["resolutions"] = resolutions
                        modes.append(cur_mode)
                device["modes"] = modes
                libcam_devs.append(device)
        return libcam_devs


class BaseProvider:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.shutdown_action = config.get("shutdown_action", "poweroff")
        self.shutdown_action = self.shutdown_action.lower()
        if self.shutdown_action not in ["halt", "poweroff"]:
            raise config.error(
                "Section [machine], Option 'shutdown_action':"
                f"Invalid value '{self.shutdown_action}', must be "
                "'halt' or 'poweroff'"
            )
        self.available_services: Dict[str, Dict[str, str]] = {}
        self.shell_cmd: SCMDComp = self.server.load_component(
            config, 'shell_command')

    async def initialize(self) -> None:
        pass

    async def _exec_sudo_command(self, command: str):
        machine: Machine = self.server.lookup_component("machine")
        return await machine.exec_sudo_command(command)

    async def shutdown(self) -> None:
        await self._exec_sudo_command(f"systemctl {self.shutdown_action}")

    async def reboot(self) -> None:
        await self._exec_sudo_command("systemctl reboot")

    async def do_service_action(self,
                                action: str,
                                service_name: str
                                ) -> None:
        raise self.server.error("Service Actions Not Available", 503)

    async def check_virt_status(self) -> Dict[str, Any]:
        return {
            'virt_type': "unknown",
            'virt_identifier': "unknown"
        }

    def is_service_available(self, service: str) -> bool:
        return service in self.available_services

    def get_available_services(self) -> Dict[str, Dict[str, str]]:
        return self.available_services

    async def extract_service_info(
        self,
        service_name: str,
        pid: int,
        properties: Optional[List[str]] = None,
        raw: bool = False
    ) -> Dict[str, Any]:
        return {}

class SystemdCliProvider(BaseProvider):
    async def initialize(self) -> None:
        await self._detect_active_services()
        if self.available_services:
            svcs = list(self.available_services.keys())
            self.svc_cmd = self.shell_cmd.build_shell_command(
                "systemctl show -p ActiveState,SubState --value "
                f"{' '.join(svcs)}")
            await self._update_service_status(0, notify=True)
            pstats: ProcStats = self.server.lookup_component('proc_stats')
            pstats.register_stat_callback(self._update_service_status)

    async def do_service_action(self,
                                action: str,
                                service_name: str
                                ) -> None:
        await self._exec_sudo_command(f"systemctl {action} {service_name}")

    async def check_virt_status(self) -> Dict[str, Any]:
        # Fallback virtualization check
        virt_id = virt_type = "none"

        # Check for any form of virtualization.  This will report the innermost
        # virtualization type in the event that nested virtualization is used
        try:
            resp: str = await self.shell_cmd.exec_cmd("systemd-detect-virt")
        except self.shell_cmd.error:
            pass
        else:
            virt_id = resp.strip()

        if virt_id != "none":
            # Check explicitly for container virtualization
            try:
                resp = await self.shell_cmd.exec_cmd(
                    "systemd-detect-virt --container")
            except self.shell_cmd.error:
                virt_type = "vm"
            else:
                if virt_id == resp.strip():
                    virt_type = "container"
                else:
                    # Moonraker is run from within a VM inside a container
                    virt_type = "vm"
            logging.info(
                f"Virtualized Environment Detected, Type: {virt_type} "
                f"id: {virt_id}")
        else:
            logging.info("No Virtualization Detected")
        return {
            'virt_type': virt_type,
            'virt_identifier': virt_id
        }

    async def _detect_active_services(self) -> None:
        machine: Machine = self.server.lookup_component("machine")
        try:
            resp: str = await self.shell_cmd.exec_cmd(
                "systemctl list-units --all --type=service --plain"
                " --no-legend")
            lines = resp.split('\n')
            services = [line.split()[0].strip() for line in lines
                        if ".service" in line.strip()]
        except Exception:
            services = []
        for svc in services:
            sname = svc.rsplit('.', 1)[0]
            if machine.is_service_allowed(sname):
                self.available_services[sname] = {
                    'active_state': "unknown",
                    'sub_state': "unknown"
                }

    async def _update_service_status(self,
                                     sequence: int,
                                     notify: bool = True
                                     ) -> None:
        if sequence % 2:
            # Update every other sequence
            return
        svcs = list(self.available_services.keys())
        try:
            resp = await self.svc_cmd.run_with_response(log_complete=False)
            for svc, state in zip(svcs, resp.strip().split('\n\n')):
                active_state, sub_state = state.split('\n', 1)
                new_state: Dict[str, str] = {
                    'active_state': active_state,
                    'sub_state': sub_state
                }
                if self.available_services[svc] != new_state:
                    self.available_services[svc] = new_state
                    if notify:
                        self.server.send_event(
                            "machine:service_state_changed",
                            {svc: new_state})
        except Exception:
            logging.exception("Error processing service state update")

    async def extract_service_info(
        self,
        service_name: str,
        pid: int,
        properties: Optional[List[str]] = None,
        raw: bool = False
    ) -> Dict[str, Any]:
        service_info: Dict[str, Any] = {}
        expected_name = f"{service_name}.service"
        if properties is None:
            properties = SERVICE_PROPERTIES
        try:
            resp: str = await self.shell_cmd.exec_cmd(
                f"systemctl status {pid}"
            )
            unit_name = resp.split(maxsplit=2)[1]
            service_info["unit_name"] = unit_name
            service_info["is_default"] = True
            service_info["manager"] = "systemd"
            if unit_name != expected_name:
                service_info["is_default"] = False
                logging.info(
                    f"Detected alternate unit name for {service_name}: "
                    f"{unit_name}"
                )
            prop_args = ",".join(properties)
            props: str = await self.shell_cmd.exec_cmd(
                f"systemctl show -p {prop_args} {unit_name}", attempts=5,
                timeout=10.
            )
            raw_props: Dict[str, Any] = {}
            lines = [p.strip() for p in props.split("\n") if p.strip()]
            for line in lines:
                parts = line.split("=", 1)
                if len(parts) == 2:
                    key = parts[0].strip()
                    val = parts[1].strip()
                    raw_props[key] = val
            if raw:
                service_info["properties"] = raw_props
            else:
                processed = self._process_raw_properties(raw_props)
                service_info["properties"] = processed
        except Exception:
            logging.exception("Error extracting service info")
            return {}
        return service_info

    def _process_raw_properties(
        self, raw_props: Dict[str, str]
    ) -> Dict[str, Any]:
        processed: Dict[str, Any] = {}
        for key, val in raw_props.items():
            processed[key] = val
            if key == "ExecStart":
                # this is a struct, we need to deconstruct it
                match = re.search(r"argv\[\]=([^;]+);", val)
                if match is not None:
                    processed[key] = match.group(1).strip()
            elif key == "EnvironmentFiles":
                if val:
                    processed[key] = val.split()[0]
            elif key in ["Requires", "After", "SupplementaryGroups"]:
                vals = [v.strip() for v in val.split() if v.strip()]
                processed[key] = vals
        return processed

class SystemdDbusProvider(BaseProvider):
    def __init__(self, config: ConfigHelper) -> None:
        super().__init__(config)
        self.dbus_mgr: DbusManager = self.server.lookup_component(
            "dbus_manager")
        self.login_mgr: Optional[ProxyInterface] = None
        self.props: List[Tuple[ProxyInterface, Callable]] = []

    async def initialize(self) -> None:
        if not self.dbus_mgr.is_connected():
            self.server.add_warning(
                "[machine]: DBus Connection Not available, systemd "
                " service tracking and actions are disabled")
            return
        # Get the systemd manager interface
        self.systemd_mgr = await self.dbus_mgr.get_interface(
            "org.freedesktop.systemd1",
            "/org/freedesktop/systemd1",
            "org.freedesktop.systemd1.Manager"
        )
        # Check for systemd PolicyKit Permissions
        await self.dbus_mgr.check_permission(
            "org.freedesktop.systemd1.manage-units",
            "System Service Management (start, stop, restart) "
            "will be disabled")
        if self.shutdown_action == "poweroff":
            await self.dbus_mgr.check_permission(
                "org.freedesktop.login1.power-off",
                "The shutdown API will be disabled"
            )
            await self.dbus_mgr.check_permission(
                "org.freedesktop.login1.power-off-multiple-sessions",
                "The shutdown API will be disabled if multiple user "
                "sessions are open."
            )
        else:
            await self.dbus_mgr.check_permission(
                "org.freedesktop.login1.halt",
                "The shutdown API will be disabled"
            )
            await self.dbus_mgr.check_permission(
                "org.freedesktop.login1.halt-multiple-sessions",
                "The shutdown API will be disabled if multiple user "
                "sessions are open."
            )
        try:
            # Get the login manaager interface
            self.login_mgr = await self.dbus_mgr.get_interface(
                "org.freedesktop.login1",
                "/org/freedesktop/login1",
                "org.freedesktop.login1.Manager"
            )
        except self.dbus_mgr.DbusError as e:
            logging.info(
                "Unable to acquire the systemd-logind D-Bus interface, "
                f"falling back to CLI Reboot and Shutdown APIs. {e}")
            self.login_mgr = None
        else:
            # Check for logind permissions
            await self.dbus_mgr.check_permission(
                "org.freedesktop.login1.reboot",
                "The reboot API will be disabled"
            )
            await self.dbus_mgr.check_permission(
                "org.freedesktop.login1.reboot-multiple-sessions",
                "The reboot API will be disabled if multiple user "
                "sessions are open."
            )
        await self._detect_active_services()

    async def reboot(self) -> None:
        if self.login_mgr is None:
            await super().reboot()
        await self.login_mgr.call_reboot(False)  # type: ignore

    async def shutdown(self) -> None:
        if self.login_mgr is None:
            await super().shutdown()
        if self.shutdown_action == "poweroff":
            await self.login_mgr.call_power_off(False)  # type: ignore
        else:
            await self.login_mgr.call_halt(False)  # type: ignore

    async def do_service_action(self,
                                action: str,
                                service_name: str
                                ) -> None:
        if not self.dbus_mgr.is_connected():
            raise self.server.error("DBus Not Connected, ", 503)
        mgr = self.systemd_mgr
        if not service_name.endswith(".service"):
            service_name += ".service"
        if action == "start":
            await mgr.call_start_unit(service_name, "replace")  # type: ignore
        elif action == "stop":
            await mgr.call_stop_unit(service_name, "replace")   # type: ignore
        elif action == "restart":
            await mgr.call_restart_unit(                        # type: ignore
                service_name, "replace")
        else:
            raise self.server.error(f"Invalid service action: {action}")

    async def check_virt_status(self) -> Dict[str, Any]:
        if not self.dbus_mgr.is_connected():
            return {
                'virt_type': "unknown",
                'virt_identifier': "unknown"
            }
        mgr = self.systemd_mgr
        virt_id = virt_type = "none"
        virt: str = await mgr.get_virtualization()  # type: ignore
        virt = virt.strip()
        if virt:
            virt_id = virt
            container_types = [
                "openvz", "lxc", "lxc-libvirt", "systemd-nspawn",
                "docker", "podman", "rkt", "wsl", "proot", "pouch"]
            if virt_id in container_types:
                virt_type = "container"
            else:
                virt_type = "vm"
            logging.info(
                f"Virtualized Environment Detected, Type: {virt_type} "
                f"id: {virt_id}")
        else:
            logging.info("No Virtualization Detected")
        return {
            'virt_type': virt_type,
            'virt_identifier': virt_id
        }

    async def _detect_active_services(self) -> None:
        # Get loaded service
        mgr = self.systemd_mgr
        machine: Machine = self.server.lookup_component("machine")
        units: List[str]
        units = await mgr.call_list_units_filtered(["loaded"])  # type: ignore
        for unit in units:
            name: str = unit[0].split('.')[0]
            if not machine.is_service_allowed(name):
                continue
            state: str = unit[3]
            substate: str = unit[4]
            dbus_path: str = unit[6]
            if name in self.available_services:
                continue
            self.available_services[name] = {
                'active_state': state,
                'sub_state': substate
            }
            # setup state monitoring
            props = await self.dbus_mgr.get_interface(
                "org.freedesktop.systemd1", dbus_path,
                "org.freedesktop.DBus.Properties"
            )
            prop_callback = self._create_properties_callback(name)
            self.props.append((props, prop_callback))
            props.on_properties_changed(  # type: ignore
                prop_callback)

    def _create_properties_callback(self, name) -> Callable:
        def prop_wrapper(dbus_obj: str,
                         changed_props: Dict[str, Variant],
                         invalid_props: Dict[str, Variant]
                         ) -> None:
            if dbus_obj != 'org.freedesktop.systemd1.Unit':
                return
            self._on_service_update(name, changed_props)
        return prop_wrapper

    def _on_service_update(self,
                           service_name: str,
                           changed_props: Dict[str, Variant]
                           ) -> None:
        if service_name not in self.available_services:
            return
        svc = self.available_services[service_name]
        notify = False
        if "ActiveState" in changed_props:
            state: str = changed_props['ActiveState'].value
            if state != svc['active_state']:
                notify = True
                svc['active_state'] = state
        if "SubState" in changed_props:
            state = changed_props['SubState'].value
            if state != svc['sub_state']:
                notify = True
                svc['sub_state'] = state
        if notify:
            self.server.send_event("machine:service_state_changed",
                                   {service_name: dict(svc)})

    async def extract_service_info(
        self,
        service_name: str,
        pid: int,
        properties: Optional[List[str]] = None,
        raw: bool = False
    ) -> Dict[str, Any]:
        if not hasattr(self, "systemd_mgr"):
            return {}
        mgr = self.systemd_mgr
        service_info: Dict[str, Any] = {}
        expected_name = f"{service_name}.service"
        if properties is None:
            properties = SERVICE_PROPERTIES
        try:
            dbus_path: str
            dbus_path = await mgr.call_get_unit_by_pid(pid)  # type: ignore
            bus = "org.freedesktop.systemd1"
            unit_intf, svc_intf = await self.dbus_mgr.get_interfaces(
                "org.freedesktop.systemd1", dbus_path,
                [f"{bus}.Unit", f"{bus}.Service"]
            )
            unit_name = await unit_intf.get_id()  # type: ignore
            service_info["unit_name"] = unit_name
            service_info["is_default"] = True
            service_info["manager"] = "systemd"
            if unit_name != expected_name:
                service_info["is_default"] = False
                logging.info(
                    f"Detected alternate unit name for {service_name}: "
                    f"{unit_name}"
                )
            raw_props: Dict[str, Any] = {}
            for key in properties:
                snake_key = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", key).lower()
                func = getattr(unit_intf, f"get_{snake_key}", None)
                if func is None:
                    func = getattr(svc_intf, f"get_{snake_key}", None)
                    if func is None:
                        continue
                val = await func()
                raw_props[key] = val
            if raw:
                service_info["properties"] = raw_props
            else:
                processed = self._process_raw_properties(raw_props)
                service_info["properties"] = processed
        except Exception:
            logging.exception("Error Extracting Service Info")
            return {}
        return service_info

    def _process_raw_properties(
        self, raw_props: Dict[str, Any]
    ) -> Dict[str, Any]:
        processed: Dict[str, Any] = {}
        for key, val in raw_props.items():
            if key == "ExecStart":
                try:
                    val = " ".join(val[0][1])
                except Exception:
                    pass
            elif key == "EnvironmentFiles":
                try:
                    val = val[0][0]
                except Exception:
                    pass
            processed[key] = val
        return processed

# for docker klipper-moonraker image multi-service managing
# since in container, all command is launched by normal user,
# sudo_cmd is not needed.
class SupervisordCliProvider(BaseProvider):
    def __init__(self, config: ConfigHelper) -> None:
        super().__init__(config)
        self.spv_conf: str = config.get("supervisord_config_path", "")

    async def initialize(self) -> None:
        await self._detect_active_services()
        keys = ' '.join(list(self.available_services.keys()))
        if self.spv_conf:
            cmd = f"supervisorctl -c {self.spv_conf} status {keys}"
        else:
            cmd = f"supervisorctl status {keys}"
        self.svc_cmd = self.shell_cmd.build_shell_command(cmd)
        await self._update_service_status(0, notify=True)
        pstats: ProcStats = self.server.lookup_component('proc_stats')
        pstats.register_stat_callback(self._update_service_status)

    async def do_service_action(
        self, action: str, service_name: str
    ) -> None:
        # slow reaction for supervisord, timeout set to 6.0
        await self._exec_supervisorctl_command(
            f"{action} {service_name}", timeout=6.
        )

    async def _exec_supervisorctl_command(
        self,
        args: str,
        tries: int = 1,
        timeout: float = 2.,
        success_codes: Optional[List[int]] = None
    ) -> str:
        if self.spv_conf:
            cmd = f"supervisorctl -c {self.spv_conf} {args}"
        else:
            cmd = f"supervisorctl {args}"
        return await self.shell_cmd.exec_cmd(
            cmd, proc_input=None, log_complete=False, attempts=tries,
            timeout=timeout, success_codes=success_codes
        )

    def _get_active_state(self, sub_state: str) -> str:
        if sub_state == "stopping":
            return "deactivating"
        elif sub_state == "running":
            return "active"
        else:
            return "inactive"

    async def _detect_active_services(self) -> None:
        machine: Machine = self.server.lookup_component("machine")
        units: Dict[str, Any] = await self._get_process_info()
        for unit, info in units.items():
            if machine.is_service_allowed(unit):
                self.available_services[unit] = {
                    'active_state': self._get_active_state(info["state"]),
                    'sub_state': info["state"]
                }

    async def _get_process_info(
        self, process_names: Optional[List[str]] = None
    ) -> Dict[str, Any]:
        units: Dict[str, Any] = {}
        cmd = "status"
        if process_names is not None:
            cmd = f"status {' '.join(process_names)}"
        try:
            resp = await self._exec_supervisorctl_command(
                cmd, timeout=6., success_codes=[0, 3]
            )
            lines = [line.strip() for line in resp.split("\n") if line.strip()]
        except Exception:
            return {}
        for line in lines:
            parts = line.split()
            name: str = parts[0]
            state: str = parts[1].lower()
            if state == "running" and len(parts) >= 6:
                units[name] = {
                    "state": state,
                    "pid": int(parts[3].rstrip(",")),
                    "uptime": parts[5]
                }
            else:
                units[name] = {
                    "state": parts[1].lower()
                }
        return units

    async def _update_service_status(self,
                                     sequence: int,
                                     notify: bool = True
                                     ) -> None:
        if sequence % 2:
            # Update every other sequence
            return
        svcs = list(self.available_services.keys())
        try:
            # slow reaction for supervisord, timeout set to 6.0
            resp = await self.svc_cmd.run_with_response(
                log_complete=False, timeout=6., success_codes=[0, 3]
            )
            resp_l = resp.strip().split("\n")  # drop lengend
            for svc, state in zip(svcs, resp_l):
                sub_state = state.split()[1].lower()
                new_state: Dict[str, str] = {
                    'active_state': self._get_active_state(sub_state),
                    'sub_state': sub_state
                }
                if self.available_services[svc] != new_state:
                    self.available_services[svc] = new_state
                    if notify:
                        self.server.send_event(
                            "machine:service_state_changed",
                            {svc: new_state})
        except Exception:
            logging.exception("Error processing service state update")

    async def _find_service_by_pid(
        self, expected_name: str, pid: int
    ) -> Dict[str, Any]:
        service_info: Dict[str, Any] = {}
        for _ in range(5):
            proc_info = await self._get_process_info(
                list(self.available_services.keys())
            )
            service_info["unit_name"] = expected_name
            service_info["is_default"] = True
            service_info["manager"] = "supervisord"
            need_retry = False
            for name, info in proc_info.items():
                if "pid" not in info:
                    need_retry |= info["state"] == "starting"
                elif info["pid"] == pid:
                    if name != expected_name:
                        service_info["unit_name"] = name
                        service_info["is_default"] = False
                        logging.info(
                            "Detected alternate unit name for "
                            f"{expected_name}: {name}"
                        )
                    return service_info
            if need_retry:
                await asyncio.sleep(1.)
            else:
                break
        return {}

    async def extract_service_info(
        self,
        service_name: str,
        pid: int,
        properties: Optional[List[str]] = None,
        raw: bool = False
    ) -> Dict[str, Any]:
        service_info = await self._find_service_by_pid(service_name, pid)
        if not service_info:
            logging.info(
                f"Unable to locate service info for {service_name}, pid: {pid}"
            )
            return {}
        # locate supervisord.conf
        if self.spv_conf:
            spv_path = pathlib.Path(self.spv_conf)
            if not spv_path.is_file():
                logging.info(
                    f"Invalid supervisord configuration file: {self.spv_conf}"
                )
                return service_info
        else:
            default_config_locations = [
                "/etc/supervisord.conf",
                "/etc/supervisor/supervisord.conf"
            ]
            for conf_path in default_config_locations:
                spv_path = pathlib.Path(conf_path)
                if spv_path.is_file():
                    break
            else:
                logging.info("Failed to locate supervisord.conf")
                return service_info
        spv_config = configparser.ConfigParser(interpolation=None)
        spv_config.read_string(spv_path.read_text())
        unit = service_info["unit_name"]
        section_name = f"program:{unit}"
        if not spv_config.has_section(section_name):
            logging.info(
                f"Unable to location supervisor section {section_name}"
            )
            return service_info
        service_info["properties"] = dict(spv_config[section_name])
        return service_info


# Install validation
INSTALL_VERSION = 1
SERVICE_VERSION = 1

SYSTEMD_UNIT = \
"""
# systemd service file for moonraker
[Unit]
Description=API Server for Klipper SV%d
Requires=network-online.target
After=network-online.target

[Install]
WantedBy=multi-user.target

[Service]
Type=simple
User=%s
SupplementaryGroups=moonraker-admin
RemainAfterExit=yes
EnvironmentFile=%s
ExecStart=%s $MOONRAKER_ARGS
Restart=always
RestartSec=10
"""  # noqa: E122

TEMPLATE_NAME = "password_request.html"

class ValidationError(Exception):
    pass

class InstallValidator:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.config = config
        self.server.load_component(config, "template")
        self.force_validation = config.getboolean("force_validation", False)
        self.sc_enabled = config.getboolean("validate_service", True)
        self.cc_enabled = config.getboolean("validate_config", True)
        app_args = self.server.get_app_args()
        self.data_path = pathlib.Path(app_args["data_path"])
        self._update_backup_path()
        self.data_path_valid = True
        self._sudo_requested = False
        self.announcement_id = ""
        self.validation_enabled = False

    def _update_backup_path(self) -> None:
        str_time = time.strftime("%Y%m%dT%H%M%SZ", time.gmtime())
        if not hasattr(self, "backup_path"):
            self.backup_path = self.data_path.joinpath(f"backup/{str_time}")
        elif not self.backup_path.exists():
            self.backup_path = self.data_path.joinpath(f"backup/{str_time}")

    async def validation_init(self) -> None:
        db: MoonrakerDatabase = self.server.lookup_component("database")
        install_ver: Optional[int] = await db.get_item(
            "moonraker", "validate_install.install_version", None
        )
        if install_ver is None:
            # skip validation for new installs
            await db.insert_item(
                "moonraker", "validate_install.install_version", INSTALL_VERSION
            )
            install_ver = INSTALL_VERSION
        if install_ver < INSTALL_VERSION:
            logging.info("Validation version in database out of date")
            self.validation_enabled = True
        else:
            msg = "Installation version in database up to date"
            if self.force_validation:
                msg += ", force is enabled"
            logging.info(msg)
            self.validation_enabled = self.force_validation
        is_bkp_cfg = self.server.get_app_args().get("is_backup_config", False)
        if self.validation_enabled and is_bkp_cfg:
            self.server.add_warning(
                "Backup configuration loaded, aborting install validation. "
                "Please correct the configuration issue and restart moonraker."
            )
            self.validation_enabled = False
            return

    async def perform_validation(self) -> bool:
        db: MoonrakerDatabase = self.server.lookup_component("database")
        if not self.validation_enabled:
            return False
        fm: FileManager = self.server.lookup_component("file_manager")
        need_restart: bool = False
        has_error: bool = False
        name = "service"
        try:
            need_restart = await self._check_service_file()
            name = "config"
            need_restart |= await self._check_configuration()
        except asyncio.CancelledError:
            raise
        except ValidationError as ve:
            has_error = True
            self.server.add_warning(str(ve))
            fm.disable_write_access()
        except Exception as e:
            has_error = True
            msg = f"Failed to validate {name}: {e}"
            self.server.add_warning(msg, exc_info=e)
            fm.disable_write_access()
        else:
            self.validation_enabled = False
            await db.insert_item(
                "moonraker", "validate_install.install_version", INSTALL_VERSION
            )
        if not has_error and need_restart:
            machine: Machine = self.server.lookup_component("machine")
            machine.restart_moonraker_service()
            return True
        return False

    async def _check_service_file(self) -> bool:
        if not self.sc_enabled:
            return False
        machine: Machine = self.server.lookup_component("machine")
        if machine.is_inside_container():
            raise ValidationError(
                "Moonraker instance running inside a container, "
                "cannot validate service file."
            )
        if machine.get_provider_type() == "none":
            raise ValidationError(
                "No machine provider configured, cannot validate service file."
            )
        logging.info("Performing Service Validation...")
        app_args = self.server.get_app_args()
        svc_info = machine.get_moonraker_service_info()
        if not svc_info:
            raise ValidationError(
                "Unable to retrieve Moonraker service info.  Service file "
                "must be updated manually."
            )
        props: Dict[str, str] = svc_info.get("properties", {})
        if "FragmentPath" not in props:
            raise ValidationError(
                "Unable to locate path to Moonraker's service unit. Service "
                "file must be must be updated manually."
            )
        desc = props.get("Description", "")
        ver_match = re.match(r"API Server for Klipper SV(\d+)", desc)
        if ver_match is not None and int(ver_match.group(1)) == SERVICE_VERSION:
            logging.info("Service file validated and up to date")
            return False
        unit: str = svc_info.get("unit_name", "").split(".", 1)[0]
        if not unit:
            raise ValidationError(
                "Unable to retrieve service unit name.  Service file "
                "must be updated manually."
            )
        if unit != "moonraker":
            logging.info(f"Custom service file detected: {unit}")
            # Not using he default unit name
            if app_args["is_default_data_path"] and self.data_path_valid:
                # No datapath set, create a new, unique data path
                df = f"~/{unit}_data"
                match = re.match(r"moonraker[-_]?(\d+)", unit)
                if match is not None:
                    df = f"~/printer_{match.group(1)}_data"
                new_dp = pathlib.Path(df).expanduser().resolve()
                if new_dp.exists() and not self._check_path_bare(new_dp):
                    raise ValidationError(
                        f"Cannot resolve data path for custom unit '{unit}', "
                        f"data path '{new_dp}' already exists.  Service file "
                        "must be updated manually."
                    )

                # If the current path is bare we can remove it
                if (
                    self.data_path.exists() and
                    self._check_path_bare(self.data_path)
                ):
                    shutil.rmtree(self.data_path)
                self.data_path = new_dp
                if not self.data_path.exists():
                    logging.info(f"New data path created at {self.data_path}")
                    self.data_path.mkdir()
                # A non-default datapath requires successful update of the
                # service
                self.data_path_valid = False
        user: str = props["User"]
        has_sudo = False
        if await machine.check_sudo_access():
            has_sudo = True
            logging.info("Moonraker has sudo access")
        elif user == "pi" and machine.sudo_password is None:
            machine.sudo_password = "raspberry"
            has_sudo = await machine.check_sudo_access()
        if not has_sudo:
            self._request_sudo_access()
            raise ValidationError(
                "Moonraker requires sudo permission to update the system "
                "service. Please check your notifications for further "
                "intructions."
            )
        self._sudo_requested = False
        svc_dest = pathlib.Path(props["FragmentPath"])
        tmp_svc = pathlib.Path(
            tempfile.gettempdir()
        ).joinpath(f"{unit}-tmp.svc")
        # Create local environment file
        sysd_data = self.data_path.joinpath("systemd")
        if not sysd_data.exists():
            sysd_data.mkdir()
        env_file = sysd_data.joinpath("moonraker.env")
        env_vars: Dict[str, str] = {
            "MOONRAKER_DATA_PATH": str(self.data_path)
        }
        cfg_file = pathlib.Path(app_args["config_file"])
        fm: FileManager = self.server.lookup_component("file_manager")
        cfg_path = fm.get_directory("config")
        log_path = fm.get_directory("logs")
        if not cfg_path or not cfg_file.parent.samefile(cfg_path):
            env_vars["MOONRAKER_CONFIG_PATH"] = str(cfg_file)
        elif cfg_file.name != "moonraker.conf":
            cfg_file = self.data_path.joinpath(f"config/{cfg_file.name}")
            env_vars["MOONRAKER_CONFIG_PATH"] = str(cfg_file)
        if not app_args["log_file"]:
            #  No log file configured
            env_vars["MOONRAKER_DISABLE_FILE_LOG"] = "y"
        else:
            # Log file does not exist in log path
            log_file = pathlib.Path(app_args["log_file"])
            if not log_path or not log_file.parent.samefile(log_path):
                env_vars["MOONRAKER_LOG_PATH"] = str(log_file)
            elif log_file.name != "moonraker.log":
                cfg_file = self.data_path.joinpath(f"logs/{log_file.name}")
                env_vars["MOONRAKER_LOG_PATH"] = str(log_file)
        # backup existing service files
        self._update_backup_path()
        svc_bkp_path = self.backup_path.joinpath("service")
        os.makedirs(str(svc_bkp_path), exist_ok=True)
        if env_file.exists():
            env_bkp = svc_bkp_path.joinpath(env_file.name)
            shutil.copy2(str(env_file), str(env_bkp))
        service_bkp = svc_bkp_path.joinpath(svc_dest.name)
        shutil.copy2(str(svc_dest), str(service_bkp))
        # write temporary service file
        src_path = source_info.source_path()
        exec_path = pathlib.Path(sys.executable)
        py_exec = exec_path.parent.joinpath("python")
        if exec_path.name == "python" or py_exec.is_file():
            # Default to loading via the python executable.  This
            # makes it possible to switch between git repos, pip
            # releases and git releases without reinstalling the
            # service.
            exec_path = py_exec
            env_vars["MOONRAKER_ARGS"] = "-m moonraker"
        if not source_info.is_dist_package():
            # This module isn't in site/dist packages,
            # add PYTHONPATH env variable
            env_vars["PYTHONPATH"] = str(src_path)
        tmp_svc.write_text(
            SYSTEMD_UNIT
            % (SERVICE_VERSION, user, env_file, exec_path)
        )
        try:
            # write new environment
            envout = "\n".join(f"{key}=\"{val}\"" for key, val in env_vars.items())
            env_file.write_text(envout)
            await machine.exec_sudo_command(
                f"cp -f {tmp_svc} {svc_dest}", tries=5, timeout=60.)
            await machine.exec_sudo_command(
                "systemctl daemon-reload", tries=5, timeout=60.
            )
        except asyncio.CancelledError:
            raise
        except Exception:
            logging.exception("Failed to update moonraker service unit")
            raise ValidationError(
                f"Failed to update service unit file '{svc_dest}'. Update must "
                f"be performed manually."
            ) from None
        finally:
            tmp_svc.unlink()
        self.data_path_valid = True
        self.sc_enabled = False
        return True

    def _check_path_bare(self, path: pathlib.Path) -> bool:
        empty: bool = True
        if not path.exists():
            return True
        for item in path.iterdir():
            if (
                item.is_file() or
                item.is_symlink() or
                item.name not in ["gcodes", "config", "logs", "certs"]
            ):
                empty = False
                break
            if item.is_dir() and next(item.iterdir(), None) is not None:
                empty = False
                break
        return empty

    def _link_data_subfolder(
        self,
        folder_name: str,
        source_dir: Union[str, pathlib.Path],
        exist_ok: bool = False
    ) -> None:
        if isinstance(source_dir, str):
            source_dir = pathlib.Path(source_dir).expanduser().resolve()
        subfolder = self.data_path.joinpath(folder_name)
        if not source_dir.exists():
            logging.info(
                f"Source path '{source_dir}' does not exist.  Falling "
                f"back to default folder {subfolder}"
            )
            return
        if not source_dir.is_dir():
            raise ValidationError(
                f"Failed to link subfolder '{folder_name}' to source path "
                f"'{source_dir}'.  The requested path is not a valid directory."
            )
        if subfolder.is_symlink():
            if not subfolder.samefile(source_dir):
                if exist_ok:
                    logging.info(
                        f"Folder {subfolder} already linked, aborting link "
                        f"to {source_dir}"
                    )
                    return
                raise ValidationError(
                    f"Failed to link subfolder '{folder_name}' to "
                    f"'{source_dir}'.  '{folder_name}' already exists and is "
                    f"linked to {subfolder}.  This conflict requires "
                    "manual resolution."
                )
            return
        if not subfolder.exists():
            subfolder.symlink_to(source_dir)
            return
        if subfolder.is_dir() and next(subfolder.iterdir(), None) is None:
            subfolder.rmdir()
            subfolder.symlink_to(source_dir)
            return
        if exist_ok:
            logging.info(
                f"Path at {subfolder} exists, aborting link to {source_dir}"
            )
            return
        raise ValidationError(
            f"Failed to link subfolder '{folder_name}' to '{source_dir}'.  "
            f"Folder '{folder_name}' already exists.  This conflict requires "
            "manual resolution."
        )

    def _link_data_file(
        self,
        data_file: Union[str, pathlib.Path],
        target: Union[str, pathlib.Path]
    ) -> None:
        if isinstance(data_file, str):
            data_file = pathlib.Path(data_file)
        if isinstance(target, str):
            target = pathlib.Path(target)
        target = target.expanduser().resolve()
        if not target.exists():
            logging.info(
                f"Target file {target} does not exist.  Aborting symbolic "
                f"link to {data_file.name}."
            )
            return
        if not target.is_file():
            raise ValidationError(
                f"Failed to link data file {data_file.name}.  Target "
                f"{target} is not a valid file."
            )
        if data_file.is_symlink():
            if not data_file.samefile(target):
                raise ValidationError(
                    f"Failed to link data file {data_file.name}.  Link "
                    f"to {data_file.resolve()} already exists. This conflict "
                    "must be resolved manually."
                )
            return
        if not data_file.exists():
            data_file.symlink_to(target)
            return
        raise ValidationError(
            f"Failed to link data file {data_file.name}.  File already exists. "
            f"This conflict must be resolved manually."
        )

    async def _check_configuration(self) -> bool:
        if not self.cc_enabled or not self.data_path_valid:
            return False
        db: MoonrakerDatabase = self.server.lookup_component("database")
        cfg_source = cast(FileSourceWrapper, self.config.get_source())
        cfg_source.backup_source()
        try:
            # write current configuration to backup path
            self._update_backup_path()
            cfg_bkp_path = self.backup_path.joinpath("config")
            os.makedirs(str(cfg_bkp_path), exist_ok=True)
            await cfg_source.write_config(cfg_bkp_path)
            # Create symbolic links for configured folders
            server_cfg = self.config["server"]

            db_cfg = self.config["database"]
            # symlink database path first
            db_path = db_cfg.get("database_path", None)
            default_db = pathlib.Path("~/.moonraker_database").expanduser()
            if db_path is None and default_db.exists():
                self._link_data_subfolder(
                    "database", default_db, exist_ok=True
                )
            elif db_path is not None:
                self._link_data_subfolder("database", db_path)
                cfg_source.remove_option("database", "database_path")

            fm_cfg = self.config["file_manager"]
            cfg_path = fm_cfg.get("config_path", None)
            if cfg_path is None:
                cfg_path = server_cfg.get("config_path", None)
            if cfg_path is not None:
                self._link_data_subfolder("config", cfg_path)
                cfg_source.remove_option("server", "config_path")
                cfg_source.remove_option("file_manager", "config_path")

            log_path = fm_cfg.get("log_path", None)
            if log_path is None:
                log_path = server_cfg.get("log_path", None)
            if log_path is not None:
                self._link_data_subfolder("logs", log_path)
                cfg_source.remove_option("server", "log_path")
                cfg_source.remove_option("file_manager", "log_path")

            gc_path: Optional[str] = await db.get_item(
                "moonraker", "file_manager.gcode_path", None
            )
            if gc_path is not None:
                self._link_data_subfolder("gcodes", gc_path)
                db.delete_item("moonraker", "file_manager.gcode_path")

            # Link individual files
            secrets_path = self.config["secrets"].get("secrets_path", None)
            if secrets_path is not None:
                secrets_dest = self.data_path.joinpath("moonraker.secrets")
                self._link_data_file(secrets_dest, secrets_path)
                cfg_source.remove_option("secrets", "secrets_path")
            certs_path = self.data_path.joinpath("certs")
            if not certs_path.exists():
                certs_path.mkdir()
            ssl_cert = server_cfg.get("ssl_certificate_path", None)
            if ssl_cert is not None:
                cert_dest = certs_path.joinpath("moonraker.cert")
                self._link_data_file(cert_dest, ssl_cert)
                cfg_source.remove_option("server", "ssl_certificate_path")
            ssl_key = server_cfg.get("ssl_key_path", None)
            if ssl_key is not None:
                key_dest = certs_path.joinpath("moonraker.key")
                self._link_data_file(key_dest, ssl_key)
                cfg_source.remove_option("server", "ssl_key_path")

            # Remove deprecated debug options
            if server_cfg.has_option("enable_debug_logging"):
                cfg_source.remove_option("server", "enable_debug_logging")
            um_cfg = server_cfg["update_manager"]
            if um_cfg.has_option("enable_repo_debug"):
                cfg_source.remove_option("update_manager", "enable_repo_debug")
        except Exception:
            cfg_source.cancel()
            raise
        finally:
            self.cc_enabled = False
        return await cfg_source.save()

    def _request_sudo_access(self) -> None:
        if self._sudo_requested:
            return
        self._sudo_requested = True
        machine: Machine = self.server.lookup_component("machine")
        machine.register_sudo_request(
            self._on_password_received,
            "Sudo password required to update Moonraker's systemd service."
        )
        if not machine.public_ip:
            async def wrapper(pub_ip):
                if not pub_ip:
                    return
                await self.remove_announcement()
                self._announce_sudo_request()
            self.server.register_event_handler(
                "machine:public_ip_changed", wrapper
            )
        self._announce_sudo_request()

    def _announce_sudo_request(self) -> None:
        machine: Machine = self.server.lookup_component("machine")
        host_info = self.server.get_host_info()
        host_addr: str = host_info["address"]
        if host_addr.lower() not in ["all", "0.0.0.0", "::"]:
            address = host_addr
        else:
            address = machine.public_ip
        if not address:
            address = f"{host_info['hostname']}.local"
        elif ":" in address:
            # ipv6 address
            address = f"[{address}]"
        app: MoonrakerApp = self.server.lookup_component("application")
        scheme = "https" if app.https_enabled() else "http"
        host_info = self.server.get_host_info()
        port = host_info["port"]
        url = f"{scheme}://{address}:{port}/"
        ancmp: Announcements = self.server.lookup_component("announcements")
        entry = ancmp.add_internal_announcement(
            "Sudo Password Required",
            "Moonraker requires sudo access to finish updating. "
            "Please click on the attached link and follow the "
            "instructions.",
            url, "high", "machine"
        )
        self.announcement_id = entry.get("entry_id", "")
        gc_announcement = (
            "!! ATTENTION: Moonraker requires sudo access to complete "
            "the update.  Go to the following URL and provide your linux "
            f"password: {url}"
        )
        self.server.send_event("server:gcode_response", gc_announcement)

    async def remove_announcement(self) -> None:
        if not self.announcement_id:
            return
        ancmp: Announcements = self.server.lookup_component("announcements")
        # remove stale announcement
        try:
            await ancmp.remove_announcement(self.announcement_id)
        except self.server.error:
            pass
        self.announcement_id = ""

    async def _on_password_received(self) -> Tuple[str, bool]:
        name = "Service"
        try:
            await self._check_service_file()
            name = "Config"
            await self._check_configuration()
        except asyncio.CancelledError:
            raise
        except Exception:
            logging.exception(f"{name} validation failed")
            raise self.server.error(
                f"{name} validation failed", 500
            ) from None
        await self.remove_announcement()
        db: MoonrakerDatabase = self.server.lookup_component("database")
        await db.insert_item(
            "moonraker", "validate_install.install_version", INSTALL_VERSION
        )
        self.validation_enabled = False
        return "System update complete.", True

def load_component(config: ConfigHelper) -> Machine:
    return Machine(config)



# Utilities for managing python packages using Pip
#
# Copyright (C) 2024 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import os
import re
import shlex
import subprocess
import pathlib
import shutil
import threading
from dataclasses import dataclass

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Union,
    Dict,
    List,
    Tuple,
    Callable,
    IO
)

if TYPE_CHECKING:
    from ..server import Server
    from ..components.shell_command import ShellCommandFactory

MIN_PIP_VERSION = (24, 0)
MIN_PYTHON_VERSION = (3, 7)

# Synchronous Subprocess Helpers
def _run_subprocess_with_response(
    cmd: str,
    timeout: Optional[float] = None,
    env: Optional[Dict[str, str]] = None
) -> str:
    prog = shlex.split(cmd)
    proc = subprocess.run(
        prog, capture_output=True, timeout=timeout, env=env,
        check=True, text=True, errors="ignore", encoding="utf-8"
    )
    if proc.returncode == 0:
        return proc.stdout.strip()
    err = proc.stderr
    raise Exception(f"Failed to run pip command '{cmd}': {err}")

def _process_subproc_output(
    stdout: IO[str],
    callback: Callable[[str], None]
) -> None:
    for line in stdout:
        callback(line.rstrip("\n"))

def _run_subprocess(
    cmd: str,
    timeout: Optional[float] = None,
    env: Optional[Dict[str, str]] = None,
    response_cb: Optional[Callable[[str], None]] = None
) -> None:
    prog = shlex.split(cmd)
    params: Dict[str, Any] = {"errors": "ignore", "encoding": "utf-8"}
    if response_cb is not None:
        params = {"stdout": subprocess.PIPE, "stderr": subprocess.STDOUT}
    with subprocess.Popen(prog, text=True, env=env, **params) as process:
        if process.stdout is not None and response_cb is not None:
            reader_thread = threading.Thread(
                target=_process_subproc_output, args=(process.stdout, response_cb)
            )
            reader_thread.start()
            reader_thread.join(timeout)
            if reader_thread.is_alive():
                process.kill()
        elif timeout is not None:
            process.wait(timeout)
    ret = process.poll()
    if ret != 0:
        raise Exception(f"Failed to run pip command '{cmd}'")

@ dataclass(frozen=True)
class PipVersionInfo:
    pip_version_string: str
    python_version_string: str

    @property
    def pip_version(self) -> Tuple[int, ...]:
        return tuple(int(part) for part in self.pip_version_string.split("."))

    @property
    def python_version(self) -> Tuple[int, ...]:
        return tuple(int(part) for part in self.python_version_string.split("."))

class PipExecutor:
    def __init__(
        self, pip_cmd: str, response_handler: Optional[Callable[[str], None]] = None
    ) -> None:
        self.pip_cmd = pip_cmd
        self.response_hdlr = response_handler

    def call_pip_with_response(
        self,
        args: str,
        timeout: Optional[float] = None,
        env: Optional[Dict[str, str]] = None
    ) -> str:
        return _run_subprocess_with_response(f"{self.pip_cmd} {args}", timeout, env)

    def call_pip(
        self,
        args: str,
        timeout: Optional[float] = None,
        env: Optional[Dict[str, str]] = None
    ) -> None:
        _run_subprocess(f"{self.pip_cmd} {args}", timeout, env, self.response_hdlr)

    def get_pip_version(self) -> PipVersionInfo:
        resp = self.call_pip_with_response("--version", 10.)
        return parse_pip_version(resp)

    def update_pip(self) -> None:
        pip_ver = ".".join([str(part) for part in MIN_PIP_VERSION])
        self.call_pip(f"install pip=={pip_ver}", 120.)

    def install_packages(
        self,
        packages: Union[pathlib.Path, List[str]],
        sys_env_vars: Optional[Dict[str, Any]] = None
    ) -> None:
        args = prepare_install_args(packages)
        env: Optional[Dict[str, str]] = None
        if sys_env_vars is not None:
            env = dict(os.environ)
            env.update(sys_env_vars)
        self.call_pip(f"install {args}", timeout=1200., env=env)

    def build_virtualenv(self, py_exec: pathlib.Path, args: str) -> None:
        bin_dir = py_exec.parent
        env_path = bin_dir.parent.resolve()
        if env_path.exists():
            shutil.rmtree(env_path)
        _run_subprocess(
            f"virtualenv {args} {env_path}",
            timeout=600.,
            response_cb=self.response_hdlr
        )
        if not py_exec.exists():
            raise Exception("Failed to create new virtualenv", 500)

class AsyncPipExecutor:
    def __init__(
        self,
        pip_cmd: str,
        server: Server,
        notify_callback: Optional[Callable[[bytes], None]] = None
    ) -> None:
        self.pip_cmd = pip_cmd
        self.server = server
        self.notify_callback = notify_callback

    def get_shell_cmd(self) -> ShellCommandFactory:
        return self.server.lookup_component("shell_command")

    async def call_pip_with_response(
        self,
        args: str,
        timeout: float = 30.,
        attempts: int = 3,
        sys_env_vars: Optional[Dict[str, Any]] = None
    ) -> str:
        env: Optional[Dict[str, str]] = None
        if sys_env_vars is not None:
            env = dict(os.environ)
            env.update(sys_env_vars)
        shell_cmd = self.get_shell_cmd()
        return await shell_cmd.exec_cmd(
            f"{self.pip_cmd} {args}", attempts=attempts,
            timeout=timeout, env=env, log_stderr=True
        )

    async def call_pip(
        self,
        args: str,
        timeout: float = 30.,
        attempts: int = 3,
        sys_env_vars: Optional[Dict[str, Any]] = None
    ) -> None:
        env: Optional[Dict[str, str]] = None
        if sys_env_vars is not None:
            env = dict(os.environ)
            env.update(sys_env_vars)
        shell_cmd = self.get_shell_cmd()
        await shell_cmd.run_cmd_async(
            f"{self.pip_cmd} {args}", self.notify_callback,
            timeout=timeout, attempts=attempts, env=env,
            log_stderr=True
        )

    async def get_pip_version(self) -> PipVersionInfo:
        resp: str = await self.get_shell_cmd().exec_cmd(
            f"{self.pip_cmd} --version", timeout=30., attempts=3, log_stderr=True
        )
        return parse_pip_version(resp)

    async def update_pip(self) -> None:
        pip_ver = ".".join([str(part) for part in MIN_PIP_VERSION])
        shell_cmd = self.get_shell_cmd()
        await shell_cmd.run_cmd_async(
            f"{self.pip_cmd} install pip=={pip_ver}",
            self.notify_callback, timeout=1200., attempts=3, log_stderr=True
        )

    async def install_packages(
        self,
        packages: Union[pathlib.Path, List[str]],
        sys_env_vars: Optional[Dict[str, Any]] = None
    ) -> None:
        # Update python dependencies
        args = prepare_install_args(packages)
        env: Optional[Dict[str, str]] = None
        if sys_env_vars is not None:
            env = dict(os.environ)
            env.update(sys_env_vars)
        shell_cmd = self.get_shell_cmd()
        await shell_cmd.run_cmd_async(
            f"{self.pip_cmd} install {args}", self.notify_callback,
            timeout=1200., attempts=3, env=env, log_stderr=True
        )

    async def build_virtualenv(self, py_exec: pathlib.Path, args: str) -> None:
        bin_dir = py_exec.parent
        env_path = bin_dir.parent.resolve()
        if env_path.exists():
            shutil.rmtree(env_path)
        shell_cmd = self.get_shell_cmd()
        await shell_cmd.exec_cmd(f"virtualenv {args} {env_path}", timeout=600.)
        if not py_exec.exists():
            raise self.server.error("Failed to create new virtualenv", 500)

def read_requirements_file(requirements_path: pathlib.Path) -> List[str]:
    if not requirements_path.is_file():
        raise FileNotFoundError(f"Requirements file {requirements_path} not found")
    data = requirements_path.read_text()
    modules: List[str] = []
    for line in data.split("\n"):
        line = line.strip()
        if not line or line[0] in "#-":
            continue
        match = re.search(r"\s#", line)
        if match is not None:
            line = line[:match.start()].strip()
        modules.append(line)
    return modules

def parse_pip_version(pip_response: str) -> PipVersionInfo:
    match = re.match(
        r"^pip ([0-9.]+) from .+? \(python ([0-9.]+)\)$", pip_response.strip()
    )
    if match is None:
        raise ValueError("Unable to parse pip version from response")
    pipver_str: str = match.group(1).strip()
    pyver_str: str = match.group(2).strip()
    return PipVersionInfo(pipver_str, pyver_str)

def check_pip_needs_update(version_info: PipVersionInfo) -> bool:
    if version_info.python_version < MIN_PYTHON_VERSION:
        return False
    return version_info.pip_version < MIN_PIP_VERSION

def prepare_install_args(packages: Union[pathlib.Path, List[str]]) -> str:
    if isinstance(packages, pathlib.Path):
        if not packages.is_file():
            raise FileNotFoundError(
                f"Invalid path to requirements_file '{packages}'"
            )
        return f"-r {packages}"
    reqs = [req.replace("\"", "'") for req in packages]
    return " ".join([f"\"{req}\"" for req in reqs])



#!/usr/bin/env python3
# Moonraker - HTTP/Websocket API Server for Klipper
#
# Copyright (C) 2020 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import pathlib
import sys
import argparse
import importlib
import os
import io
import time
import socket
import logging
import signal
import asyncio
import uuid
import traceback
from . import confighelper
from .eventloop import EventLoop
from .utils import (
    ServerError,
    Sentinel,
    get_software_info,
    json_wrapper,
    pip_utils,
    source_info
)
from .loghelper import LogManager
from .common import RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Callable,
    Coroutine,
    Dict,
    List,
    Tuple,
    Union,
    TypeVar,
)
if TYPE_CHECKING:
    from .common import WebRequest
    from .components.application import MoonrakerApp
    from .components.websockets import WebsocketManager
    from .components.klippy_connection import KlippyConnection
    from .components.file_manager.file_manager import FileManager
    from .components.machine import Machine
    from .components.extensions import ExtensionManager
    FlexCallback = Callable[..., Optional[Coroutine]]
    _T = TypeVar("_T", Sentinel, Any)

API_VERSION = (1, 4, 0)
SERVER_COMPONENTS = ['application', 'websockets', 'klippy_connection']
CORE_COMPONENTS = [
    'dbus_manager', 'database', 'file_manager', 'authorization',
    'klippy_apis', 'machine', 'data_store', 'shell_command',
    'proc_stats', 'job_state', 'job_queue', 'history',
    'http_client', 'announcements', 'webcam', 'extensions'
]


class Server:
    error = ServerError
    config_error = confighelper.ConfigError
    def __init__(self,
                 args: Dict[str, Any],
                 log_manager: LogManager,
                 event_loop: EventLoop
                 ) -> None:
        self.event_loop = event_loop
        self.log_manager = log_manager
        self.app_args = args
        self.events: Dict[str, List[FlexCallback]] = {}
        self.components: Dict[str, Any] = {}
        self.failed_components: List[str] = []
        self.warnings: Dict[str, str] = {}
        self._is_configured: bool = False

        self.config = config = self._parse_config()
        self.host: str = config.get('host', "0.0.0.0")
        self.port: int = config.getint('port', 7125)
        self.ssl_port: int = config.getint('ssl_port', 7130)
        self.exit_reason: str = ""
        self.server_running: bool = False
        self.app_running_evt = asyncio.Event()
        self.pip_recovery_attempted: bool = False

        # Configure Debug Logging
        config.getboolean('enable_debug_logging', False, deprecate=True)
        self.debug = args["debug"]
        log_level = logging.DEBUG if args["verbose"] else logging.INFO
        logging.getLogger().setLevel(log_level)
        self.event_loop.set_debug(args["asyncio_debug"])
        self.klippy_connection: KlippyConnection
        self.klippy_connection = self.load_component(config, "klippy_connection")

        # Tornado Application/Server
        self.moonraker_app: MoonrakerApp = self.load_component(config, "application")
        app = self.moonraker_app
        self.register_endpoint = app.register_endpoint
        self.register_debug_endpoint = app.register_debug_endpoint
        self.register_static_file_handler = app.register_static_file_handler
        self.register_upload_handler = app.register_upload_handler
        self.log_manager.set_server(self)
        self.websocket_manager: WebsocketManager
        self.websocket_manager = self.load_component(config, "websockets")

        for warning in args.get("startup_warnings", []):
            self.add_warning(warning)

        self.register_endpoint(
            "/server/info", RequestType.GET, self._handle_info_request
        )
        self.register_endpoint(
            "/server/config", RequestType.GET, self._handle_config_request
        )
        self.register_endpoint(
            "/server/restart", RequestType.POST, self._handle_server_restart
        )
        self.register_notification("server:klippy_ready")
        self.register_notification("server:klippy_shutdown")
        self.register_notification("server:klippy_disconnect",
                                   "klippy_disconnected")
        self.register_notification("server:gcode_response")

    def get_app_args(self) -> Dict[str, Any]:
        return dict(self.app_args)

    def get_app_arg(self, key: str, default=Sentinel.MISSING) -> Any:
        val = self.app_args.get(key, default)
        if val is Sentinel.MISSING:
            raise KeyError(f"No key '{key}' in Application Arguments")
        return val

    def get_event_loop(self) -> EventLoop:
        return self.event_loop

    def get_api_version(self) -> Tuple[int, int, int]:
        return API_VERSION

    def get_warnings(self) -> List[str]:
        return list(self.warnings.values())

    def is_running(self) -> bool:
        return self.server_running

    def is_configured(self) -> bool:
        return self._is_configured

    def is_debug_enabled(self) -> bool:
        return self.debug

    def is_verbose_enabled(self) -> bool:
        return self.app_args["verbose"]

    def _parse_config(self) -> confighelper.ConfigHelper:
        config = confighelper.get_configuration(self, self.app_args)
        # log config file
        cfg_files = "\n".join(config.get_config_files())
        strio = io.StringIO()
        config.write_config(strio)
        cfg_item = f"\n{'#'*20} Moonraker Configuration {'#'*20}\n\n"
        cfg_item += strio.getvalue()
        cfg_item += "#"*65
        cfg_item += f"\nAll Configuration Files:\n{cfg_files}\n"
        cfg_item += "#"*65
        strio.close()
        self.add_log_rollover_item('config', cfg_item)
        return config

    async def server_init(self, start_server: bool = True) -> None:
        self.event_loop.add_signal_handler(
            signal.SIGTERM, self._handle_term_signal)

        # Perform asynchronous init after the event loop starts
        optional_comps: List[Coroutine] = []
        for name, component in self.components.items():
            if not hasattr(component, "component_init"):
                continue
            if name in CORE_COMPONENTS:
                # Process core components in order synchronously
                await self._initialize_component(name, component)
            else:
                optional_comps.append(
                    self._initialize_component(name, component))

        # Asynchronous Optional Component Initialization
        if optional_comps:
            await asyncio.gather(*optional_comps)

        if not self.warnings:
            await self.event_loop.run_in_thread(self.config.create_backup)

        machine: Machine = self.lookup_component("machine")
        restarting = await machine.validate_installation()
        if not restarting and start_server:
            await self.start_server()

    async def start_server(self, connect_to_klippy: bool = True) -> None:
        # Open Unix Socket Server
        extm: ExtensionManager = self.lookup_component("extensions")
        await extm.start_unix_server()

        # Start HTTP Server
        logging.info(
            f"Starting Moonraker on ({self.host}, {self.port}), "
            f"Hostname: {socket.gethostname()}")
        self.moonraker_app.listen(self.host, self.port, self.ssl_port)
        self.server_running = True
        if connect_to_klippy:
            self.klippy_connection.connect()

    async def run_until_exit(self) -> None:
        await self.app_running_evt.wait()

    def add_log_rollover_item(
        self, name: str, item: str, log: bool = True
    ) -> None:
        self.log_manager.set_rollover_info(name, item)
        if log and item is not None:
            logging.info(item)

    def add_warning(
        self,
        warning: str,
        warn_id: Optional[str] = None,
        log: bool = True,
        exc_info: Optional[BaseException] = None
    ) -> str:
        if warn_id is None:
            warn_id = str(id(warning))
        self.warnings[warn_id] = warning
        if log:
            logging.warning(warning, exc_info=exc_info)
        return warn_id

    def remove_warning(self, warn_id: str) -> None:
        self.warnings.pop(warn_id, None)

    # ***** Component Management *****
    async def _initialize_component(self, name: str, component: Any) -> None:
        logging.info(f"Performing Component Post Init: [{name}]")
        try:
            ret = component.component_init()
            if ret is not None:
                await ret
        except Exception as e:
            self.add_warning(
                f"Component '{name}' failed to load with error: {e}", exc_info=e
            )
            self.set_failed_component(name)

    def load_components(self) -> None:
        config = self.config
        cfg_sections = set([s.split()[0] for s in config.sections()])
        cfg_sections.remove('server')

        # load database to initialize saved state
        self.load_component(config, "database")
        self.klippy_connection.load_saved_state()

        # load core components
        for component in CORE_COMPONENTS:
            self.load_component(config, component)
            if component in cfg_sections:
                cfg_sections.remove(component)

        # load remaining optional components
        for section in cfg_sections:
            self.load_component(config, section, None)

        config.validate_config()
        self._is_configured = True

    def load_component(
        self,
        config: confighelper.ConfigHelper,
        component_name: str,
        default: _T = Sentinel.MISSING
    ) -> Union[_T, Any]:
        if component_name in self.components:
            return self.components[component_name]
        if self.is_configured():
            raise self.error(
                "Cannot load components after configuration", 500
            )
        if component_name in self.failed_components:
            raise self.error(
                f"Component {component_name} previously failed to load", 500
            )
        try:
            full_name = f"moonraker.components.{component_name}"
            module = importlib.import_module(full_name)
            # Server components use the [server] section for configuration
            if component_name not in SERVER_COMPONENTS:
                is_core = component_name in CORE_COMPONENTS
                fallback: Optional[str] = "server" if is_core else None
                config = config.getsection(component_name, fallback)
            load_func = getattr(module, "load_component")
            component = load_func(config)
        except Exception as e:
            ucomps: List[str] = self.app_args.get("unofficial_components", [])
            if isinstance(e, ModuleNotFoundError) and component_name not in ucomps:
                if self.try_pip_recovery(e.name or "unknown"):
                    return self.load_component(config, component_name, default)
            msg = f"Unable to load component: ({component_name})"
            logging.exception(msg)
            if component_name not in self.failed_components:
                self.failed_components.append(component_name)
            if default is Sentinel.MISSING:
                raise
            return default
        self.components[component_name] = component
        logging.info(f"Component ({component_name}) loaded")
        return component

    def try_pip_recovery(self, missing_module: str) -> bool:
        if self.pip_recovery_attempted:
            return False
        self.pip_recovery_attempted = True
        src_dir = source_info.source_path()
        req_file = src_dir.joinpath("scripts/moonraker-requirements.txt")
        if not req_file.is_file():
            return False
        pip_cmd = f"{sys.executable} -m pip"
        pip_exec = pip_utils.PipExecutor(pip_cmd, logging.info)
        logging.info(f"Module '{missing_module}' not found. Attempting Pip Update...")
        logging.info("Checking Pip Version...")
        try:
            pipver = pip_exec.get_pip_version()
            if pip_utils.check_pip_needs_update(pipver):
                cur_ver = pipver.pip_version_string
                new_ver = ".".join([str(part) for part in pip_utils.MIN_PIP_VERSION])
                logging.info(f"Updating Pip from {cur_ver} to {new_ver}...")
                pip_exec.update_pip()
        except Exception:
            logging.exception("Pip version check failed")
            return False
        logging.info("Installing Moonraker python dependencies...")
        try:
            pip_exec.install_packages(req_file, {"SKIP_CYTHON": "Y"})
        except Exception:
            logging.exception("Failed to install python packages")
            return False
        return True

    def lookup_component(
        self, component_name: str, default: _T = Sentinel.MISSING
    ) -> Union[_T, Any]:
        component = self.components.get(component_name, default)
        if component is Sentinel.MISSING:
            raise ServerError(f"Component ({component_name}) not found")
        return component

    def set_failed_component(self, component_name: str) -> None:
        if component_name not in self.failed_components:
            self.failed_components.append(component_name)

    def register_component(self, component_name: str, component: Any) -> None:
        if component_name in self.components:
            raise self.error(
                f"Component '{component_name}' already registered")
        self.components[component_name] = component

    def register_notification(
        self, event_name: str, notify_name: Optional[str] = None
    ) -> None:
        self.websocket_manager.register_notification(event_name, notify_name)

    def register_event_handler(
        self, event: str, callback: FlexCallback
    ) -> None:
        self.events.setdefault(event, []).append(callback)

    def send_event(self, event: str, *args) -> asyncio.Future:
        fut = self.event_loop.create_future()
        self.event_loop.register_callback(
            self._process_event, fut, event, *args)
        return fut

    async def _process_event(
        self, fut: asyncio.Future, event: str, *args
    ) -> None:
        events = self.events.get(event, [])
        coroutines: List[Coroutine] = []
        for func in events:
            try:
                ret = func(*args)
            except Exception:
                logging.exception(f"Error processing callback in event {event}")
            else:
                if ret is not None:
                    coroutines.append(ret)
        if coroutines:
            results = await asyncio.gather(*coroutines, return_exceptions=True)
            for val in results:
                if isinstance(val, Exception):
                    if sys.version_info < (3, 10):
                        exc_info = "".join(traceback.format_exception(
                            type(val), val, val.__traceback__
                        ))
                    else:
                        exc_info = "".join(traceback.format_exception(val))
                    logging.info(
                        f"\nError processing callback in event {event}\n{exc_info}"
                    )
        if not fut.done():
            fut.set_result(None)

    def register_remote_method(
        self, method_name: str, cb: FlexCallback
    ) -> None:
        self.klippy_connection.register_remote_method(method_name, cb)

    def get_host_info(self) -> Dict[str, Any]:
        return {
            'hostname': socket.gethostname(),
            'address': self.host,
            'port': self.port,
            'ssl_port': self.ssl_port
        }

    def get_klippy_info(self) -> Dict[str, Any]:
        return self.klippy_connection.klippy_info

    def _handle_term_signal(self) -> None:
        logging.info("Exiting with signal SIGTERM")
        self.event_loop.register_callback(self._stop_server, "terminate")

    def restart(self, delay: Optional[float] = None) -> None:
        if delay is None:
            self.event_loop.register_callback(self._stop_server)
        else:
            self.event_loop.delay_callback(delay, self._stop_server)

    async def _stop_server(self, exit_reason: str = "restart") -> None:
        self.server_running = False
        # Call each component's "on_exit" method
        for name, component in self.components.items():
            if hasattr(component, "on_exit"):
                func: FlexCallback = getattr(component, "on_exit")
                try:
                    ret = func()
                    if ret is not None:
                        await ret
                except Exception:
                    logging.exception(
                        f"Error executing 'on_exit()' for component: {name}")

        # Sleep for 100ms to allow connected websockets to write out
        # remaining data
        await asyncio.sleep(.1)
        try:
            await self.moonraker_app.close()
            await self.websocket_manager.close()
        except Exception:
            logging.exception("Error Closing App")

        # Disconnect from Klippy
        try:
            await asyncio.wait_for(
                asyncio.shield(self.klippy_connection.close(
                    wait_closed=True)), 2.)
        except Exception:
            logging.exception("Klippy Disconnect Error")

        # Close all components
        for name, component in self.components.items():
            if name in ["application", "websockets", "klippy_connection"]:
                # These components have already been closed
                continue
            if hasattr(component, "close"):
                func = getattr(component, "close")
                try:
                    ret = func()
                    if ret is not None:
                        await ret
                except Exception:
                    logging.exception(
                        f"Error executing 'close()' for component: {name}")
        # Allow cancelled tasks a chance to run in the eventloop
        await asyncio.sleep(.001)

        self.exit_reason = exit_reason
        self.event_loop.remove_signal_handler(signal.SIGTERM)
        self.app_running_evt.set()

    async def _handle_server_restart(self, web_request: WebRequest) -> str:
        self.event_loop.register_callback(self._stop_server)
        return "ok"

    async def _handle_info_request(self, web_request: WebRequest) -> Dict[str, Any]:
        raw = web_request.get_boolean("raw", False)
        file_manager: Optional[FileManager] = self.lookup_component(
            'file_manager', None)
        reg_dirs = []
        if file_manager is not None:
            reg_dirs = file_manager.get_registered_dirs()
        mreqs = self.klippy_connection.missing_requirements
        if raw:
            warnings = list(self.warnings.values())
        else:
            warnings = [
                w.replace("\n", "<br/>") for w in self.warnings.values()
            ]
        return {
            'klippy_connected': self.klippy_connection.is_connected(),
            'klippy_state': str(self.klippy_connection.state),
            'components': list(self.components.keys()),
            'failed_components': self.failed_components,
            'registered_directories': reg_dirs,
            'warnings': warnings,
            'websocket_count': self.websocket_manager.get_count(),
            'moonraker_version': self.app_args['software_version'],
            'missing_klippy_requirements': mreqs,
            'api_version': API_VERSION,
            'api_version_string': ".".join([str(v) for v in API_VERSION])
        }

    async def _handle_config_request(self, web_request: WebRequest) -> Dict[str, Any]:
        cfg_file_list: List[Dict[str, Any]] = []
        cfg_parent = pathlib.Path(
            self.app_args["config_file"]
        ).expanduser().resolve().parent
        for fname, sections in self.config.get_file_sections().items():
            path = pathlib.Path(fname)
            try:
                rel_path = str(path.relative_to(str(cfg_parent)))
            except ValueError:
                rel_path = fname
            cfg_file_list.append({"filename": rel_path, "sections": sections})
        return {
            'config': self.config.get_parsed_config(),
            'orig': self.config.get_orig_config(),
            'files': cfg_file_list
        }

async def launch_server(
    log_manager: LogManager, app_args: Dict[str, Any]
) -> Optional[int]:
    eventloop = EventLoop()
    startup_warnings: List[str] = app_args["startup_warnings"]
    try:
        server = Server(app_args, log_manager, eventloop)
        server.load_components()
    except confighelper.ConfigError as e:
        logging.exception("Server Config Error")
        backup_cfg: Optional[str] = app_args["backup_config"]
        if app_args["is_backup_config"] or backup_cfg is None:
            return 1
        app_args["is_backup_config"] = True
        startup_warnings.append(
            f"Server configuration error: {e}\n"
            f"Loading most recent working configuration: '{backup_cfg}'\n"
            f"Please fix the issue in moonraker.conf and restart the server."
        )
        return True
    except Exception:
        logging.exception("Moonraker Error")
        return 1
    try:
        await server.server_init()
        await server.run_until_exit()
    except Exception:
        logging.exception("Server Running Error")
        return 1
    if server.exit_reason == "terminate":
        return 0
    # Restore the original config and clear the warning
    # before the server restarts
    if app_args["is_backup_config"]:
        startup_warnings.pop()
        app_args["is_backup_config"] = False
    del server
    return None

def main(from_package: bool = True) -> None:
    def get_env_bool(key: str) -> bool:
        return os.getenv(key, "").lower() in ["y", "yes", "true"]

    # Parse start arguments
    parser = argparse.ArgumentParser(
        description="Moonraker - Klipper API Server")
    parser.add_argument(
        "-d", "--datapath",
        default=os.getenv("MOONRAKER_DATA_PATH"),
        metavar='<data path>',
        help="Location of Moonraker Data File Path"
    )
    parser.add_argument(
        "-c", "--configfile",
        default=os.getenv("MOONRAKER_CONFIG_PATH"),
        metavar='<configfile>',
        help="Path to Moonraker's configuration file"
    )
    parser.add_argument(
        "-l", "--logfile",
        default=os.getenv("MOONRAKER_LOG_PATH"),
        metavar='<logfile>',
        help="Path to Moonraker's log file"
    )
    parser.add_argument(
        "-u", "--unixsocket",
        default=os.getenv("MOONRAKER_UDS_PATH"),
        metavar="<unixsocket>",
        help="Path to Moonraker's unix domain socket"
    )
    parser.add_argument(
        "-n", "--nologfile",
        action='store_const',
        const=True,
        default=get_env_bool("MOONRAKER_DISABLE_FILE_LOG"),
        help="disable logging to a file"
    )
    parser.add_argument(
        "-v", "--verbose",
        action='store_const',
        const=True,
        default=get_env_bool("MOONRAKER_VERBOSE_LOGGING"),
        help="Enable verbose logging"
    )
    parser.add_argument(
        "-g", "--debug",
        action='store_const',
        const=True,
        default=get_env_bool("MOONRAKER_ENABLE_DEBUG"),
        help="Enable Moonraker debug features"
    )
    parser.add_argument(
        "-o", "--asyncio-debug",
        action='store_const',
        const=True,
        default=get_env_bool("MOONRAKER_ASYNCIO_DEBUG"),
        help="Enable asyncio debug flag"
    )
    cmd_line_args = parser.parse_args()

    startup_warnings: List[str] = []
    dp: str = cmd_line_args.datapath or "~/printer_data"
    data_path = pathlib.Path(dp).expanduser().resolve()
    if not data_path.exists():
        try:
            data_path.mkdir()
        except Exception:
            startup_warnings.append(
                f"Unable to create data path folder at {data_path}"
            )
    uuid_path = data_path.joinpath(".moonraker.uuid")
    if not uuid_path.is_file():
        instance_uuid = uuid.uuid4().hex
        uuid_path.write_text(instance_uuid)
    else:
        instance_uuid = uuid_path.read_text().strip()
    if cmd_line_args.configfile is not None:
        cfg_file: str = cmd_line_args.configfile
    else:
        cfg_file = str(data_path.joinpath("config/moonraker.conf"))
    if cmd_line_args.unixsocket is not None:
        unix_sock: str = cmd_line_args.unixsocket
    else:
        comms_dir = data_path.joinpath("comms")
        if not comms_dir.exists():
            comms_dir.mkdir()
        unix_sock = str(comms_dir.joinpath("moonraker.sock"))
    misc_dir = data_path.joinpath("misc")
    if not misc_dir.exists():
        misc_dir.mkdir()
    app_args = {
        "data_path": str(data_path),
        "is_default_data_path": cmd_line_args.datapath is None,
        "config_file": cfg_file,
        "backup_config": confighelper.find_config_backup(cfg_file),
        "startup_warnings": startup_warnings,
        "verbose": cmd_line_args.verbose,
        "debug": cmd_line_args.debug,
        "asyncio_debug": cmd_line_args.asyncio_debug,
        "is_backup_config": False,
        "is_python_package": from_package,
        "instance_uuid": instance_uuid,
        "unix_socket_path": unix_sock
    }

    # Setup Logging
    app_args.update(get_software_info())
    if cmd_line_args.nologfile:
        app_args["log_file"] = ""
    elif cmd_line_args.logfile:
        app_args["log_file"] = os.path.normpath(
            os.path.expanduser(cmd_line_args.logfile))
    else:
        app_args["log_file"] = str(data_path.joinpath("logs/moonraker.log"))
    app_args["python_version"] = sys.version.replace("\n", " ")
    app_args["launch_args"] = " ".join([sys.executable] + sys.argv).strip()
    app_args["msgspec_enabled"] = json_wrapper.MSGSPEC_ENABLED
    app_args["uvloop_enabled"] = EventLoop.UVLOOP_ENABLED
    log_manager = LogManager(app_args, startup_warnings)

    # Start asyncio event loop and server
    while True:
        estatus = asyncio.run(launch_server(log_manager, app_args))
        if estatus is not None:
            break
        # Since we are running outside of the the server
        # it is ok to use a blocking sleep here
        time.sleep(.5)
        logging.info("Attempting Server Restart...")
    logging.info("Server Shutdown")
    log_manager.stop_logging()
    exit(estatus)



# Utilities for enumerating devices using sysfs
#
# Copyright (C) 2024 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license
from __future__ import annotations
import os
import fcntl
import ctypes
import pathlib
import enum
from ..common import ExtendedFlag
from . import ioctl_macros
from typing import (
    Dict,
    List,
    Any,
    Union,
    Optional
)

DEFAULT_USB_IDS_PATH = "/usr/share/misc/usb.ids"
USB_DEVICE_PATH = "/sys/bus/usb/devices"
TTY_PATH = "/sys/class/tty"
SER_BYPTH_PATH = "/dev/serial/by-path"
SER_BYID_PATH = "/dev/serial/by-id"
V4L_DEVICE_PATH = "/sys/class/video4linux"
V4L_BYPTH_PATH = "/dev/v4l/by-path"
V4L_BYID_PATH = "/dev/v4l/by-id"

OPTIONAL_USB_INFO = ["manufacturer", "product", "serial"]
NULL_DESCRIPTIONS = [
    "?", "none", "undefined", "reserved/undefined", "unused", "no subclass"
]

def read_item(parent: pathlib.Path, filename: str) -> str:
    return parent.joinpath(filename).read_text().strip()

def find_usb_folder(usb_path: pathlib.Path) -> Optional[str]:
    # Find the sysfs usb folder from a child folder
    while usb_path.is_dir() and usb_path.name:
        dnum_file = usb_path.joinpath("devnum")
        bnum_file = usb_path.joinpath("busnum")
        if not dnum_file.is_file() or not bnum_file.is_file():
            usb_path = usb_path.parent
            continue
        devnum = int(dnum_file.read_text().strip())
        busnum = int(bnum_file.read_text().strip())
        return f"{busnum}:{devnum}"
    return None

class UsbIdData:
    _usb_info_cache: Dict[str, str] = {
        "DI:1d50": "OpenMoko, Inc",
        "DI:1d50:614e": "Klipper 3d-Printer Firmware",
        "DI:1d50:6177": "Katapult Bootloader (CDC_ACM)"
    }

    def __init__(self, usb_id_path: Union[str, pathlib.Path]) -> None:
        if isinstance(usb_id_path, str):
            usb_id_path = pathlib.Path(usb_id_path)
        self.usb_id_path = usb_id_path.expanduser().resolve()
        self.parsed: bool = False
        self.usb_info: Dict[str, str] = {}

    def _is_hex(self, item: str) -> bool:
        try:
            int(item, 16)
        except ValueError:
            return False
        return True

    def get_item(self, key: str, check_null: bool = False) -> Optional[str]:
        item = self.usb_info.get(key, self._usb_info_cache.get(key))
        if item is None:
            if self.parsed:
                return None
            self.parse_usb_ids()
            item = self.usb_info.get(key)
            if item is None:
                return None
        self._usb_info_cache[key] = item
        if check_null and item.lower() in NULL_DESCRIPTIONS:
            return None
        return item

    def parse_usb_ids(self) -> None:
        self.parsed = True
        if not self.usb_id_path.is_file():
            return
        top_key: str = ""
        sub_key: str = ""
        with self.usb_id_path.open(encoding="latin-1") as f:
            while True:
                line = f.readline()
                if not line:
                    break
                stripped_line = line.strip()
                if not stripped_line or stripped_line[0] == "#":
                    continue
                if line[:2] == "\t\t":
                    if not sub_key:
                        continue
                    tertiary_id, desc = stripped_line.split(maxsplit=1)
                    self.usb_info[f"{sub_key}:{tertiary_id.lower()}"] = desc
                elif line[0] == "\t":
                    if not top_key:
                        continue
                    sub_id, desc = stripped_line.split(maxsplit=1)
                    sub_key = f"{top_key}:{sub_id.lower()}"
                    self.usb_info[sub_key] = desc
                else:
                    id_type, data = line.rstrip().split(maxsplit=1)
                    if len(id_type) == 4 and self._is_hex(id_type):
                        # This is a vendor ID
                        top_key = f"DI:{id_type.lower()}"
                        self.usb_info[top_key] = data
                    elif id_type:
                        # This is a subtype
                        num_id, desc = data.split(maxsplit=1)
                        top_key = f"{id_type}:{num_id.lower()}"
                        self.usb_info[top_key] = desc
                    else:
                        break

    def get_product_info(self, vendor_id: str, product_id: str) -> Dict[str, Any]:
        vendor_name = self.get_item(f"DI:{vendor_id}")
        if vendor_name is None:
            return {
                "description": None,
                "manufacturer": None,
                "product": None,
            }
        product_name = self.get_item(f"DI:{vendor_id}:{product_id}")
        return {
            "description": f"{vendor_name} {product_name or ''}".strip(),
            "manufacturer": vendor_name,
            "product": product_name,
        }

    def get_class_info(
        self, cls_id: str, subcls_id: str, proto_id: str
    ) -> Dict[str, Any]:
        cls_desc = self.get_item(f"C:{cls_id}")
        if cls_desc is None or cls_id == "00":
            return {
                "class": None,
                "subclass": None,
                "protocol": None
            }
        return {
            "class": cls_desc,
            "subclass": self.get_item(f"C:{cls_id}:{subcls_id}", True),
            "protocol": self.get_item(f"C:{cls_id}:{subcls_id}:{proto_id}", True)
        }

def find_usb_devices() -> List[Dict[str, Any]]:
    dev_folder = pathlib.Path(USB_DEVICE_PATH)
    if not dev_folder.is_dir():
        return []
    usb_devs: List[Dict[str, Any]] = []
    # Find sysfs usb device descriptors
    for dev_cfg_path in dev_folder.glob("*/bDeviceClass"):
        dev_folder = dev_cfg_path.parent
        device_info: Dict[str, Any] = {}
        try:
            device_info["device_num"] = int(read_item(dev_folder, "devnum"))
            device_info["bus_num"] = int(read_item(dev_folder, "busnum"))
            device_info["vendor_id"] = read_item(dev_folder, "idVendor").lower()
            device_info["product_id"] = read_item(dev_folder, "idProduct").lower()
            usb_location = f"{device_info['bus_num']}:{device_info['device_num']}"
            device_info["usb_location"] = usb_location
            dev_cls = read_item(dev_folder, "bDeviceClass").lower()
            dev_subcls = read_item(dev_folder, "bDeviceSubClass").lower()
            dev_proto = read_item(dev_folder, "bDeviceProtocol").lower()
            device_info["class_ids"] = [dev_cls, dev_subcls, dev_proto]
            for field in OPTIONAL_USB_INFO:
                if dev_folder.joinpath(field).is_file():
                    device_info[field] = read_item(dev_folder, field)
                elif field not in device_info:
                    device_info[field] = None
        except Exception:
            continue
        usb_devs.append(device_info)
    return usb_devs

def find_serial_devices() -> List[Dict[str, Any]]:
    serial_devs: List[Dict[str, Any]] = []
    devs_by_path: Dict[str, str] = {}
    devs_by_id: Dict[str, str] = {}
    by_path_dir = pathlib.Path(SER_BYPTH_PATH)
    by_id_dir = pathlib.Path(SER_BYID_PATH)
    dev_root_folder = pathlib.Path("/dev")
    if by_path_dir.is_dir():
        devs_by_path = {
            dev.resolve().name: str(dev) for dev in by_path_dir.iterdir()
        }
    if by_id_dir.is_dir():
        devs_by_id = {
            dev.resolve().name: str(dev) for dev in by_id_dir.iterdir()
        }
    tty_dir = pathlib.Path(TTY_PATH)
    for tty_path in tty_dir.iterdir():
        device_folder = tty_path.joinpath("device")
        if not device_folder.is_dir():
            continue
        uartclk_file = tty_path.joinpath("uartclk")
        port_file = tty_path.joinpath("port")
        device_name = tty_path.name
        driver_name = device_folder.joinpath("driver").resolve().name
        device_info: Dict[str, Any] = {
            "device_type": "unknown",
            "device_path": str(dev_root_folder.joinpath(device_name)),
            "device_name": device_name,
            "driver_name": driver_name,
            "path_by_hardware": devs_by_path.get(device_name),
            "path_by_id": devs_by_id.get(device_name),
            "usb_location": None
        }
        if uartclk_file.is_file() and port_file.is_file():
            # This is a potential hardware uart.  Need to
            # validate that "serial8250" devices have a port
            # number of zero
            if driver_name == "serial8250":
                portnum = int(port_file.read_text().strip(), 16)
                if portnum != 0:
                    # Not a usable UART
                    continue
            device_info["device_type"] = "hardware_uart"
        else:
            usb_path = device_folder.resolve()
            usb_location: Optional[str] = find_usb_folder(usb_path)
            if usb_location is not None:
                device_info["device_type"] = "usb"
                device_info["usb_location"] = usb_location
        serial_devs.append(device_info)
    return serial_devs

class struct_v4l2_capability(ctypes.Structure):
    _fields_ = [
        ("driver", ctypes.c_char * 16),
        ("card", ctypes.c_char * 32),
        ("bus_info", ctypes.c_char * 32),
        ("version", ctypes.c_uint32),
        ("capabilities", ctypes.c_uint32),
        ("device_caps", ctypes.c_uint32),
        ("reserved", ctypes.c_uint32 * 3),
    ]

class struct_v4l2_fmtdesc(ctypes.Structure):
    _fields_ = [
        ("index", ctypes.c_uint32),
        ("type", ctypes.c_uint32),
        ("flags", ctypes.c_uint32),
        ("description", ctypes.c_char * 32),
        ("pixelformat", ctypes.c_uint32),
        ("reserved", ctypes.c_uint32 * 4)
    ]

class struct_v4l2_frmsize_discrete(ctypes.Structure):
    _fields_ = [
        ("width", ctypes.c_uint32),
        ("height", ctypes.c_uint32),
    ]


class struct_v4l2_frmsize_stepwise(ctypes.Structure):
    _fields_ = [
        ("min_width", ctypes.c_uint32),
        ("max_width", ctypes.c_uint32),
        ("step_width", ctypes.c_uint32),
        ("min_height", ctypes.c_uint32),
        ("max_height", ctypes.c_uint32),
        ("step_height", ctypes.c_uint32),
    ]

class struct_v4l2_frmsize_union(ctypes.Union):
    _fields_ = [
        ("discrete", struct_v4l2_frmsize_discrete),
        ("stepwise", struct_v4l2_frmsize_stepwise)
    ]

class struct_v4l2_frmsizeenum(ctypes.Structure):
    _anonymous_ = ("size",)
    _fields_ = [
        ("index", ctypes.c_uint32),
        ("pixel_format", ctypes.c_uint32),
        ("type", ctypes.c_uint32),
        ("size", struct_v4l2_frmsize_union),
        ("reserved", ctypes.c_uint32 * 2)
    ]

class V4L2Capability(ExtendedFlag):
    VIDEO_CAPTURE        = 0x00000001  # noqa: E221
    VIDEO_OUTPUT         = 0x00000002  # noqa: E221
    VIDEO_OVERLAY        = 0x00000004  # noqa: E221
    VBI_CAPTURE          = 0x00000010  # noqa: E221
    VBI_OUTPUT           = 0x00000020  # noqa: E221
    SLICED_VBI_CAPTURE   = 0x00000040  # noqa: E221
    SLICED_VBI_OUTPUT    = 0x00000080  # noqa: E221
    RDS_CAPTURE          = 0x00000100  # noqa: E221
    VIDEO_OUTPUT_OVERLAY = 0x00000200
    HW_FREQ_SEEK         = 0x00000400  # noqa: E221
    RDS_OUTPUT           = 0x00000800  # noqa: E221
    VIDEO_CAPTURE_MPLANE = 0x00001000
    VIDEO_OUTPUT_MPLANE  = 0x00002000  # noqa: E221
    VIDEO_M2M_MPLANE     = 0x00004000  # noqa: E221
    VIDEO_M2M            = 0x00008000  # noqa: E221
    TUNER                = 0x00010000  # noqa: E221
    AUDIO                = 0x00020000  # noqa: E221
    RADIO                = 0x00040000  # noqa: E221
    MODULATOR            = 0x00080000  # noqa: E221
    SDR_CAPTURE          = 0x00100000  # noqa: E221
    EXT_PIX_FORMAT       = 0x00200000  # noqa: E221
    SDR_OUTPUT           = 0x00400000  # noqa: E221
    META_CAPTURE         = 0x00800000  # noqa: E221
    READWRITE            = 0x01000000  # noqa: E221
    STREAMING            = 0x04000000  # noqa: E221
    META_OUTPUT          = 0x08000000  # noqa: E221
    TOUCH                = 0x10000000  # noqa: E221
    IO_MC                = 0x20000000  # noqa: E221
    SET_DEVICE_CAPS      = 0x80000000  # noqa: E221

class V4L2FrameSizeTypes(enum.IntEnum):
    DISCRETE = 1
    CONTINUOUS = 2
    STEPWISE = 3

class V4L2FormatFlags(ExtendedFlag):
    COMPRESSED = 0x0001
    EMULATED = 0x0002


V4L2_BUF_TYPE_VIDEO_CAPTURE = 1
V4L2_QUERYCAP = ioctl_macros.IOR(ord("V"), 0, struct_v4l2_capability)
V4L2_ENUM_FMT = ioctl_macros.IOWR(ord("V"), 2, struct_v4l2_fmtdesc)
V4L2_ENUM_FRAMESIZES = ioctl_macros.IOWR(ord("V"), 74, struct_v4l2_frmsizeenum)

def v4l2_fourcc_from_fmt(pixelformat: int) -> str:
    fmt = bytes([((pixelformat >> (8 * i)) & 0xFF) for i in range(4)])
    return fmt.decode(encoding="ascii", errors="ignore")

def v4l2_fourcc(format: str) -> int:
    assert len(format) == 4
    result: int = 0
    for idx, val in enumerate(format.encode()):
        result |= (val << (8 * idx)) & 0xFF
    return result

def _get_resolutions(fd: int, pixel_format: int) -> List[str]:
    res_info = struct_v4l2_frmsizeenum()
    result: List[str] = []
    for idx in range(128):
        res_info.index = idx
        res_info.pixel_format = pixel_format
        try:
            fcntl.ioctl(fd, V4L2_ENUM_FRAMESIZES, res_info)
        except OSError:
            break
        if res_info.type != V4L2FrameSizeTypes.DISCRETE:
            break
        width = res_info.discrete.width
        height = res_info.discrete.height
        result.append(f"{width}x{height}")
    return result

def _get_modes(fd: int) -> List[Dict[str, Any]]:
    pix_info = struct_v4l2_fmtdesc()
    result: List[Dict[str, Any]] = []
    for idx in range(128):
        pix_info.index = idx
        pix_info.type = V4L2_BUF_TYPE_VIDEO_CAPTURE
        try:
            fcntl.ioctl(fd, V4L2_ENUM_FMT, pix_info)
        except OSError:
            break
        desc: str = pix_info.description.decode()
        pixel_format: int = pix_info.pixelformat
        flags = V4L2FormatFlags(pix_info.flags)
        resolutions = _get_resolutions(fd, pixel_format)
        if not resolutions:
            continue
        result.append(
            {
                "format": v4l2_fourcc_from_fmt(pixel_format),
                "description": desc,
                "flags": [f.name for f in flags],
                "resolutions": resolutions
            }
        )
    return result

def find_video_devices() -> List[Dict[str, Any]]:
    v4lpath = pathlib.Path(V4L_DEVICE_PATH)
    if not v4lpath.is_dir():
        return []
    v4l_by_path_dir = pathlib.Path(V4L_BYPTH_PATH)
    v4l_by_id_dir = pathlib.Path(V4L_BYID_PATH)
    dev_root_folder = pathlib.Path("/dev")
    v4l_devs_by_path: Dict[str, str] = {}
    v4l_devs_by_id: Dict[str, str] = {}
    if v4l_by_path_dir.is_dir():
        v4l_devs_by_path = {
            dev.resolve().name: str(dev) for dev in v4l_by_path_dir.iterdir()
        }
    if v4l_by_id_dir.is_dir():
        v4l_devs_by_id = {
            dev.resolve().name: str(dev) for dev in v4l_by_id_dir.iterdir()
        }
    v4l_devices: List[Dict[str, Any]] = []
    for v4ldev_path in v4lpath.iterdir():
        devfs_name = v4ldev_path.name
        devfs_path = dev_root_folder.joinpath(devfs_name)
        # The video4linux sysfs implmentation provides limited device
        # info.  Use the VIDEOC_QUERYCAPS ioctl to retreive extended
        # information about the v4l2 device.
        fd: int = -1
        try:
            fd = os.open(str(devfs_path), os.O_RDONLY | os.O_NONBLOCK)
            cap_info = struct_v4l2_capability()
            fcntl.ioctl(fd, V4L2_QUERYCAP, cap_info)
            capabilities = V4L2Capability(cap_info.device_caps)
            if not capabilities & V4L2Capability.VIDEO_CAPTURE:
                # Skip devices that do not capture video
                continue
            modes = _get_modes(fd)
        except Exception:
            continue
        finally:
            if fd != -1:
                os.close(fd)
        ver_tuple = tuple(
            [str((cap_info.version >> (i)) & 0xFF) for i in range(16, -1, -8)]
        )
        video_device: Dict[str, Any] = {
            "device_name": devfs_name,
            "device_path": str(devfs_path),
            "camera_name": cap_info.card.decode(),
            "driver_name": cap_info.driver.decode(),
            "hardware_bus": cap_info.bus_info.decode(),
            "capabilities": [cap.name for cap in capabilities],
            "version": ".".join(ver_tuple),
            "path_by_hardware": v4l_devs_by_path.get(devfs_name),
            "path_by_id": v4l_devs_by_id.get(devfs_name),
            "alt_name": None,
            "usb_location": None,
            "modes": modes
        }
        name_file = v4ldev_path.joinpath("name")
        if name_file.is_file():
            video_device["alt_name"] = read_item(v4ldev_path, "name")
        device_path = v4ldev_path.joinpath("device")
        if device_path.is_dir():
            usb_location = find_usb_folder(device_path.resolve())
            if usb_location is not None:
                video_device["usb_location"] = usb_location
        v4l_devices.append(video_device)

    def idx_sorter(item: Dict[str, Any]) -> int:
        try:
            return int(item["device_name"][5:])
        except ValueError:
            return -1
    # Sort by string first, then index
    v4l_devices.sort(key=lambda item: item["device_name"])
    v4l_devices.sort(key=idx_sorter)
    return v4l_devices



# General Server Utilities
#
# Copyright (C) 2020 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import logging
import os
import glob
import importlib
import pathlib
import sys
import subprocess
import asyncio
import hashlib
import shlex
import re
import struct
import socket
import enum
import ipaddress
import platform
from .exceptions import ServerError
from . import source_info
from . import json_wrapper

# Annotation imports
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
    Any,
    Tuple,
    Dict,
    Union
)

if TYPE_CHECKING:
    from types import ModuleType
    from asyncio.trsock import TransportSocket

SYS_MOD_PATHS = glob.glob("/usr/lib/python3*/dist-packages")
SYS_MOD_PATHS += glob.glob("/usr/lib/python3*/site-packages")
SYS_MOD_PATHS += glob.glob("/usr/lib/*-linux-gnu/python3*/site-packages")
IPAddress = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]

try:
    KERNEL_VERSION = tuple([int(part) for part in platform.release().split(".")[:2]])
except Exception:
    KERNEL_VERSION = (0, 0)


class Sentinel(enum.Enum):
    MISSING = object()

def _run_git_command(cmd: str) -> str:
    prog = shlex.split(cmd)
    process = subprocess.Popen(prog, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    ret, err = process.communicate()
    retcode = process.wait()
    if retcode == 0:
        return ret.strip().decode()
    raise Exception(
        f"Failed to run git command '{cmd}': {err.decode(errors='ignore')}"
    )

def _retrieve_git_tag(source_path: str) -> str:
    cmd = f"git -C {source_path} rev-list --tags --max-count=1"
    hash = _run_git_command(cmd)
    cmd = f"git -C {source_path} describe --tags {hash}"
    tag = _run_git_command(cmd)
    cmd = f"git -C {source_path} rev-list {tag}..HEAD --count"
    count = _run_git_command(cmd)
    return f"{tag}-{count}"

# Parse the git version from the command line.  This code
# is borrowed from Klipper.
def retrieve_git_version(source_path: str) -> str:
    # Obtain version info from "git" program
    cmd = f"git -C {source_path} describe --always --tags --long --dirty"
    ver = _run_git_command(cmd)
    tag_match = re.match(r"v\d+\.\d+\.\d+", ver)
    if tag_match is not None:
        return ver
    # This is likely a shallow clone.  Resolve the tag and manually create
    # the version string
    tag = _retrieve_git_tag(source_path)
    return f"t{tag}-g{ver}-shallow"

def get_repo_info(source_path: str) -> Dict[str, Any]:
    repo_info: Dict[str, Any] = {
        "software_version": "?",
        "git_branch": "?",
        "git_remote": "?",
        "git_repo_url": "?",
        "modified_files": [],
        "unofficial_components": []
    }
    try:
        repo_info["software_version"] = retrieve_git_version(source_path)
        cmd = f"git -C {source_path} branch --no-color"
        branch_list = _run_git_command(cmd)
        for line in branch_list.split("\n"):
            if line[0] == "*":
                repo_info["git_branch"] = line[1:].strip()
                break
        else:
            return repo_info
        if repo_info["git_branch"].startswith("(HEAD detached"):
            parts = repo_info["git_branch"] .strip("()").split()[-1]
            remote, _, _ = parts.partition("/")
            if not remote:
                return repo_info
            repo_info["git_remote"] = remote
        else:
            branch = repo_info["git_branch"]
            cmd = f"git -C {source_path} config --get branch.{branch}.remote"
            repo_info["git_remote"] = _run_git_command(cmd)
        cmd = f"git -C {source_path} remote get-url {repo_info['git_remote']}"
        repo_info["git_repo_url"] = _run_git_command(cmd)
        cmd = f"git -C {source_path} status --porcelain --ignored"
        status = _run_git_command(cmd)
        for line in status.split("\n"):
            parts = line.strip().split(maxsplit=1)
            if len(parts) != 2:
                continue
            if parts[0] == "M":
                repo_info["modified_files"].append(parts[1])
            elif (
                parts[0] in ("??", "!!")
                and parts[1].endswith(".py")
                and parts[1].startswith("components")
            ):
                comp = parts[1].split("/", maxsplit=1)[-1]
                repo_info["unofficial_components"].append(comp)
    except Exception:
        logging.exception("Error Retreiving Git Repo Info")
    return repo_info

def get_software_info() -> Dict[str, Any]:
    src_path = source_info.source_path()
    if source_info.is_git_repo():
        return get_repo_info(str(src_path))
    pkg_ver = source_info.package_version()
    if pkg_ver is not None:
        return {"software_version": pkg_ver}
    version: str = "?"
    vfile = src_path.joinpath("moonraker/.version")
    if vfile.exists():
        try:
            version = vfile.read_text().strip()
        except Exception:
            logging.exception("Unable to extract version from file")
            version = "?"
    return {"software_version": version}

def hash_directory(
    dir_path: Union[str, pathlib.Path],
    ignore_exts: List[str],
    ignore_dirs: List[str]
) -> str:
    if isinstance(dir_path, str):
        dir_path = pathlib.Path(dir_path)
    checksum = hashlib.blake2s()
    if not dir_path.exists():
        return ""
    for dpath, dnames, fnames in os.walk(dir_path):
        valid_dirs: List[str] = []
        for dname in sorted(dnames):
            if dname[0] == '.' or dname in ignore_dirs:
                continue
            valid_dirs.append(dname)
        dnames[:] = valid_dirs
        for fname in sorted(fnames):
            ext = os.path.splitext(fname)[-1].lower()
            if fname[0] == '.' or ext in ignore_exts:
                continue
            fpath = pathlib.Path(os.path.join(dpath, fname))
            try:
                checksum.update(fpath.read_bytes())
            except Exception:
                pass
    return checksum.hexdigest()

def verify_source(
    path: Optional[Union[str, pathlib.Path]] = None
) -> Optional[Tuple[str, bool]]:
    if path is None:
        path = source_info.source_path()
    elif isinstance(path, str):
        path = pathlib.Path(path)
    rfile = path.joinpath(".release_info")
    if not rfile.exists():
        return None
    try:
        rinfo = json_wrapper.loads(rfile.read_text())
    except Exception:
        return None
    orig_chksum = rinfo['source_checksum']
    ign_dirs = rinfo['ignored_dirs']
    ign_exts = rinfo['ignored_exts']
    checksum = hash_directory(path, ign_exts, ign_dirs)
    return checksum, checksum == orig_chksum

def load_system_module(name: str) -> ModuleType:
    if not SYS_MOD_PATHS:
        # no dist path detected, fall back to direct import attempt
        try:
            return importlib.import_module(name)
        except ImportError as e:
            raise ServerError(f"Unable to import module {name}") from e
    for module_path in SYS_MOD_PATHS:
        sys.path.insert(0, module_path)
        try:
            module = importlib.import_module(name)
        except ImportError as e:
            if not isinstance(e, ModuleNotFoundError):
                logging.exception(f"Failed to load {name} module")
        else:
            break
        finally:
            sys.path.pop(0)
    else:
        raise ServerError(f"Unable to import module {name}")
    return module

def get_unix_peer_credentials(
    writer: asyncio.StreamWriter, name: str
) -> Dict[str, int]:
    sock: TransportSocket
    sock = writer.get_extra_info("socket", None)
    if sock is None:
        logging.debug(
            f"Unable to get underlying Unix Socket for {name}, "
            "cant fetch peer credentials"
        )
        return {}
    data: bytes = b""
    try:
        size = struct.calcsize("3I")
        data = sock.getsockopt(socket.SOL_SOCKET, socket.SO_PEERCRED, size)
        pid, uid, gid = struct.unpack("3I", data)
    except asyncio.CancelledError:
        raise
    except Exception:
        logging.exception(
            f"Failed to get Unix Socket Peer Credentials for {name}"
            f", raw: 0x{data.hex()}"
        )
        return {}
    return {
        "process_id": pid,
        "user_id": uid,
        "group_id": gid
    }

def pretty_print_time(seconds: int) -> str:
    if seconds == 0:
        return "0 Seconds"
    fmt_list: List[str] = []
    times: Dict[str, int] = {}
    times["Day"], seconds = divmod(seconds, 86400)
    times["Hour"], seconds = divmod(seconds, 3600)
    times["Minute"], times["Second"] = divmod(seconds, 60)
    for ident, val in times.items():
        if val == 0:
            continue
        fmt_list.append(f"{val} {ident}" if val == 1 else f"{val} {ident}s")
    return ", ".join(fmt_list)

def parse_ip_address(address: str) -> Optional[IPAddress]:
    try:
        return ipaddress.ip_address(address)
    except Exception:
        return None
